"""
Contains the base Telescope class.
"""

import azcam
from azcam.header import Header, ObjectHeaderMethods
from azcam.tools.tools import Tools


class Telescope(Tools, ObjectHeaderMethods):
    """
    The base telescope tool.
    Usually implemented as the "telescope" tool.
    """

    def __init__(self, tool_id="telescope", description=None):
        Tools.__init__(self, tool_id, description)

        # focus position
        self.focus_position = 0

        # create the temp control Header object
        self.header = Header("Telescope")
        self.header.set_header("telescope", 5)

        azcam.db.tools_init["telescope"] = self
        azcam.db.tools_reset["telescope"] = self

    def initialize(self):
        """
        Initialize tool.
        """

        self.is_initialized = 1

        return

    def reset(self):
        """
        Reset tool.
        """

        self.is_reset = 1

        return

    # ***************************************************************************
    # exposure
    # ***************************************************************************

    def exposure_start(self):
        """
        Optional call before exposure starts.
        """

        return

    def exposure_finish(self):
        """
        Optional call after exposure finishes.
        """

        return



"""
Contains the ControllerMag class.
"""

import os
import time

import azcam
import azcam.exceptions
from azcam.tools.controller import Controller

from .camera_server import CameraServerInterface

"""
FPGA system: Fill image buffer command

fillbuffer 0 val	-> fill buffer with fixed val
fillbuffer 1 		-> fill buffer with ramp 0-65535
fillbuffer 2 		-> fill buffer with ramp 0-lastcol
fillbuffer 3 		-> fill buffer with random values
"""


class ControllerMag(Controller):
    """
    The controller class for Magellan controllers.
    """

    def __init__(self, tool_id="controller", description=None):
        super().__init__(tool_id, description)

        # selected video channel(s) for switched systems
        self.video_select = 0
        # video gain flag
        self.video_gain = 1
        # True to generate fake ramp data instead of real data
        self.synthetic_data = 0

        # True if image data from ControllserServer is binary (not a file)
        self.binary_image_data = 1

        # True to use read lock in ControllerServer
        self.use_read_lock = 0

        self.controller_class = "mag"
        self.controller_type = "mag1"

        # DSP code filename
        self.timing_file = ""

        self.camserver = CameraServerInterface()

    def reset(self):
        """
        Reset controller.
        May warn that the controller could not be reset.
        """

        # reset flag even is system has previously been reset
        self.is_reset = 0

        # close and open PCI interface every time
        self.camserver.command("CloseInterface")  # don't stop if error here

        azcam.log("Resetting interface")
        self.camserver.command("OpenInterface mag1")

        self.header.delete_all_keywords()

        azcam.log("Waiting for DSP reset")
        time.sleep(1.6)

        azcam.log("Loading DSP file %s" % self.timing_file)

        try:
            self.upload_dsp_file(2, self.timing_file)
        except azcam.exceptions.AzcamError as e:
            # warn about reset
            if e.error_code == 1:
                azcam.exceptions.warning("Controller not reset: check power")
                return

        self.remove_read_lock()

        # wait for FPGA reset (104 ms)
        time.sleep(0.1)

        self.is_reset = 1

        self.set_roi()

        self.set_exposuretime(azcam.db.tools["exposure"].exposure_time)

        self.set_read_lock()

        try:
            self.camserver.command("set debug 0")
        except Exception:
            pass

        return

    def initialize(self):
        """
        Initialize the controller interface.
        """

        reply = self.camserver.get("ControllerType")
        if reply[0] == "OK":
            if reply[1] == "0" or reply[1].lower() == "demo":
                azcam.log("ControllerServer running in DEMO mode")
            self.is_initialized = True
            return
        else:
            raise azcam.exceptions.AzcamError("Could not initialize controller")

        return

    def set_read_lock(self, Flag=-1):
        """
        Set the read lock in the ControllerServer.
        If not specified then the default value is used.
        """

        if Flag == -1:
            Flag = self.use_read_lock
        else:
            self.use_read_lock = Flag  # save for future resets

        self.camserver.set("UseReadLock", Flag)

        return

    def get_read_lock(self):
        """
        Returns the read lock value in the ControllerServer.
        """

        reply = self.camserver.get("UseReadLock")
        return int(reply[1])

    def remove_read_lock(self):
        """
        Removes the read lock value in the ControllerServer.
        Should not be required.
        """

        self.camserver.command("RemoveReadLock")

        return

    # *** ROI shifting parameters  ***

    def set_roi(self):
        """
        Sets the ROI parameters values in the controller based on focalplane parameters.
        Sends parameters to the controller.
        """
        # send parameters to controller in order to do all hardware communication here
        if self.is_reset:
            cmd = (
                "SetFormat "
                + str(self.detpars.ns_total)
                + " "
                + str(self.detpars.ns_predark)
                + " "
                + str(self.detpars.ns_underscan)
                + " "
                + str(self.detpars.ns_overscan)
                + " "
                + str(self.detpars.np_total)
                + " "
                + str(self.detpars.np_predark)
                + " "
                + str(self.detpars.np_underscan)
                + " "
                + str(self.detpars.np_overscan)
                + " "
                + str(self.detpars.np_frametransfer)
            )
            self.camserver.command(cmd)

            self.camserver.command(
                "SetRoi "
                + str(self.detpars.first_col)
                + " "
                + str(self.detpars.last_col)
                + " "
                + str(self.detpars.first_row)
                + " "
                + str(self.detpars.last_row)
                + " "
                + str(self.detpars.col_bin)
                + " "
                + str(self.detpars.row_bin)
            )

        # update ControllerServer for image size
        if self.is_reset:
            self.camserver.set("NumberPixelsImage", self.detpars.numpix_image)

        return

    def flush(self, Cycles=1):
        """
        Flush or clear out the detector 'Cycles' times.
        Returns after clearing is finished which could take many seconds.
        """

        for _ in range(Cycles):
            reply = self.magio("flush_ccd", 0)

        return

    def set_shutter_state(self, Flag: bool = 0):
        """
        Sets the shutter state during an exposure.
        Flag is:
        open -> close shutter during exposure
        close -> open shutter during exposure
        """

        self.camserver.set("ShutterState", Flag)

        return

    # *** testing ***

    def test(self, Loops=10):
        """
        Perform MAG controller tests.
        Loops is the number of times the test cycle is executed.
        """

        # reset
        azcam.log("Resetting controller")
        self.reset()

        # communications
        azcam.log("Testing communication")
        for loop in range(Loops):
            azcam.log("Loop %d" % loop)
            self.test_datalink(2, 98, 25)

        return

    def test_datalink(self, BoardNumber=2, value=999, Loops=1):
        """
        Test communications to controller.
        BoardNumber - board number to test - ignored!
        value - integer number to send to board.
        Loops is the number of times to repeat command.
        """

        for _ in range(Loops):
            status = self.magio("dsp_echo", value)
            if status[0] != "OK":
                raise azcam.exceptions.AzcamError(
                    "Comminication with controller failed"
                )

        return

    # *** shutter *

    def set_shutter(self, state):
        """
        Open or close controller shutter.
        """

        if state:
            self.magio("set_shutter", 1)
        else:
            self.magio("set_shutter", 0)

        return

    # *** files ***

    def upload_dsp_file(self, BoardNumber, filename):
        """
        Sends DSP a file to a controllercontaining DSP code to the PCI, timing, or utility boards.
        """

        csfile = self.upload_file(filename)

        # wait for reset
        time.sleep(1)

        self.load_file(BoardNumber, csfile)

        # set keyword for file loaded
        if BoardNumber == 2:
            self.header.set_keyword(
                "DSPFILE",
                os.path.basename(filename),
                "Timing board DSP code filename",
                "str",
            )

        return

    def load_file(self, BoardNumber, filename):
        """
        Write a file containing DSP code to the controller board.
        BoardNumber - board number - ignored!
        filename - filename on server of DSP code.  S-record.
        """

        self.camserver.load_file(BoardNumber, filename)
        self.header.set_keyword(
            "DSPFILE", os.path.basename(filename), "DSP code filename", "str"
        )

        return

    # *** exposure time ***

    def set_exposuretime(self, ExposureTime):
        """
        Write the exposure time (in seconds) to the controller.
        """

        ExposureTime = float(ExposureTime)

        et_msec = int(ExposureTime * 1000)

        return self.camserver.set("ExposureTime", et_msec)

    def get_exposuretime(self):
        """
        Return the exposure time from the controller (in seconds).
        """

        reply = self.camserver.get("ExposureTime")  # this in msec

        return float(reply[1]) / 1000.0

    def update_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        reply = self.camserver.get("ExposureTimeRemaining")
        elapsed = int(reply[1])  # milliseconds
        print(elapsed)
        return (
            max(0, azcam.db.tools["exposure"].exposure_time * 1000 - elapsed) / 1000.0
        )

    # *** readout ***

    def start_readout(self):
        """
        Start readout immediately.
        Returns imemdiately, does not wait for readout to finish.
        """

        return self.read_image()

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        reply = self.camserver.get("PixelCount")
        count = int(reply[1])
        return max(0, self.detpars.numpix_image - count)

    # *** misc ***

    def magio(self, Command, Parameter):
        """
        Issue a low-level MAG controller IO on the controller server.
        """
        return self.camserver.command("MagIO " + str(Command) + " " + str(Parameter))

    def start_exposure(self):
        """
        Start exposure (integration).
        Returns immediately, not waiting for exposure to finish.
        """

        return self.camserver.command("StartExposure")

    def exposure_abort(self):
        """
        Abort an integration.
        """
        return self.camserver.command("AbortExposure")

    def readout_abort(self):
        """
        Abort a readout in progress.
        """
        return self.camserver.command("AbortReadout")

    def exposure_pause(self):
        """
        Pause an integration which is in progress.
        """
        return self.camserver.command("PauseExposure")

    def read_image(self):
        """
        Start readout of detector.
        """
        return self.camserver.command("ReadImage")

    def exposure_resume(self):
        """
        Resume a paused integration.
        """
        return self.camserver.command("ResumeExposure")

    def write_image(self, flag, file):
        """
        Write image to local disk on controller server.
        """

        return self.camserver.command("WriteImage " + str(flag) + " " + file)

    def set_configuration(self, Flag, Splits, numdet_x, numdet_y, amp_cfg):
        """
        Set detector configuration.
        """

        return self.camserver.command(
            "SetConfiguration "
            + str(Flag)
            + " "
            + str(Splits)
            + " "
            + str(numdet_x)
            + " "
            + str(numdet_y)
            + " "
            + str(amp_cfg)
        )

    # *** files ***

    def upload_file(self, filename):
        """
        Sends a local file to the controller server to be written to its file system.
        Returns uploaded filename on controller server.
        """

        with open(filename, "r") as f:
            fbuffer = f.read()

        # send file as binary, reply is filename on controller server
        reply = self.camserver.upload_file(fbuffer)

        return reply



import os
import socket
import time

import azcam
import azcam.exceptions
from azcam.tools.tools import Tools


class SendImage(object):
    """
    Class to send image to a remote image server.
    """

    def __init__(self):

        self.remote_imageserver_host = ""
        self.remote_imageserver_port = 0
        self.remote_imageserver_type = ""
        self.remote_imageserver_filename = ""

        self.timeout = 10.0
        self.overwrite = 0
        self.test_image = 0
        self.display_image = 0
        self.filetype = 0
        self.size_x = 0
        self.size_y = 0

        # new - imageserver call method, may be overridden
        self.imageserver_send = self.dataserver

    def set_remote_imageserver(
        self,
        remote_imageserver_host="",
        remote_imageserver_port=6543,
        remote_imageserver_type="dataserver",
        remote_imageserver_filename="image",
    ):
        """
        Set parameters so image files are sent to a remote image server.
        """

        self.remote_imageserver_host = remote_imageserver_host
        self.remote_imageserver_port = remote_imageserver_port
        self.remote_imageserver_type = remote_imageserver_type
        self.remote_imageserver_filename = remote_imageserver_filename

        if self.remote_imageserver_type == "dataserver":
            self.imageserver_send = self.dataserver
        elif self.remote_imageserver_type == "azcam":
            self.imageserver_send = self.azcam_imageserver

        return

    def get_remote_imageserver(self):
        """
        Get remote image server parameters.
        Returns:
            remote_imageserver_host:
            remote_imageserver_port:
            remote_imageserver_type:
            remote_imageserver_filename:
        """

        return [
            self.remote_imageserver_host,
            self.remote_imageserver_port,
            self.remote_imageserver_type,
            self.remote_imageserver_filename,
        ]

    def send_image(self, localfile=None, remotefile=None):
        """
        Send image to remote image server.
        """

        if localfile is None:
            localfile = f"{azcam.db.tools['exposure'].temp_image_file}.{azcam.db.tools['exposure'].get_extname(self.filetype)}"

        if remotefile is None:
            remotefile = f"{self.remote_imageserver_filename}.{azcam.db.tools['exposure'].get_extname(self.filetype)}"

        self.overwrite = azcam.db.tools["exposure"].overwrite
        self.test_image = azcam.db.tools["exposure"].test_image
        self.display_image = azcam.db.tools["exposure"].display_image
        self.filetype = azcam.db.tools["exposure"].filetype
        self.size_x = azcam.db.tools["exposure"].size_x
        self.size_y = azcam.db.tools["exposure"].size_y

        self.imageserver_send(localfile, remotefile)

        # if self.remote_imageserver_type == "azcam":
        #     self.azcam_imageserver(localfile, remotefile)
        # elif self.remote_imageserver_type == "lbtguider":
        #     self.lbtguider_imageserver(localfile, remotefile)
        # elif self.remote_imageserver_type == "dataserver":
        #     self.dataserver(localfile, remotefile)
        # elif self.remote_imageserver_type == "ccdacq":
        #     self.ccdacq_imageserver(localfile, remotefile)
        # else:
        #     raise azcam.exceptions.AzcamError("Unknown remote image server type")

    def azcam_imageserver(self, localfile, remotefile=None):
        """
        Send image to azcam image server.
        """

        if remotefile is None:
            remotefile = self.remote_imageserver_filename

        # open image file on disk
        with open(localfile, "rb") as dfile:
            lSize = os.path.getsize(localfile)
            buff = dfile.read()

        # open socket to DataServer
        dataserver_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        dataserver_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)
        dataserver_socket.settimeout(self.timeout)
        dataserver_socket.connect(
            (self.remote_imageserver_host, int(self.remote_imageserver_port))
        )

        if self.overwrite or self.test_image:
            remotefile = "!" + remotefile

        # send header
        # file types: 0 FITS, 1 MEF, 2 binary
        s1 = "%16d %s %d %d %d %d" % (
            lSize,
            remotefile,
            self.filetype,
            self.size_x,
            self.size_y,
            self.display_image,
        )
        s1 = "%-256s" % s1
        status = dataserver_socket.send(str.encode(s1))
        if status != 256:
            raise azcam.exceptions.AzcamError(
                "Could not send image header to remote image server"
            )

        # get 16 char ASCII header return status from image server
        # reply = dataserver_socket.recv(16)
        # if len(reply) != 16:
        #    raise azcam.exceptions.AzcamError("Did not receive header return status from remote image server")
        # retstat=int(reply[:1])
        # retstat=int(reply)
        retstat = 0

        # check header return status codes
        if retstat != 0:
            if retstat == -1:
                raise azcam.exceptions.AzcamError("Bad reply from remote image server")
            elif retstat == -2:
                raise azcam.exceptions.AzcamError(
                    "Remote ImageServer not create image filename"
                )
            elif retstat == -3:  #
                raise azcam.exceptions.AzcamError(
                    "Folder does not exist on remote machine"
                )
            else:
                raise azcam.exceptions.AzcamError(
                    "Unknown error from remote image server"
                )

        # send file data
        reply = dataserver_socket.send(buff)

        if reply != len(buff):
            raise azcam.exceptions.AzcamError(
                "Did not send entire image file data to remote image server"
            )

        # get 16 char ASCII final return status from image server
        reply = dataserver_socket.recv(16).decode()

        # if len(reply) != 2:
        #    raise AzcamError("did not receive entire return status from remote image server")

        retstat = int(reply[:1])

        # check final return status error codes
        if retstat != 0:
            raise azcam.exceptions.AzcamError(
                "Bad final return status from remote image server"
            )

        # close socket
        # time.sleep()
        dataserver_socket.close()

        return

    def dataserver(self, localfile, remotefile):
        """
        Send image to dataserver.
        """

        ImageSendBufferSize = 1024 * 32

        # open image file on disk
        with open(localfile, "rb") as dfile:
            lSize = os.path.getsize(localfile)
            buff = dfile.read()

        # open socket to DataServer
        dataserver_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        dataserver_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)

        try:
            dataserver_socket.settimeout(self.timeout)
            dataserver_socket.connect(
                (self.remote_imageserver_host, int(self.remote_imageserver_port))
            )
        except Exception as e:
            azcam.log(e)
            raise azcam.exceptions.AzcamError(
                f"Could not connect to imageserver {self.remote_imageserver_host}:{int(self.remote_imageserver_port)}"
            )

        if self.overwrite or self.test_image:
            remotefile = "!" + remotefile

        azcam.log(f"Sending image to {self.remote_imageserver_host} as {remotefile}")

        # send header
        # file types: 0 FITS, 1 MEF, 2 binary
        s1 = "%16d %s %d %d %d %d" % (
            lSize,
            remotefile,
            self.filetype,
            self.size_x,
            self.size_y,
            self.display_image,
        )
        s1 = "%-256s" % s1
        status = dataserver_socket.send(str.encode(s1))
        if status != 256:
            raise azcam.exceptions.AzcamError(
                "Could not send image header to remote DataServer"
            )

        # get 16 char ASCII header return status from image server
        # reply = dataserver_socket.recv(16)
        # if len(reply) != 16:
        #    raise azcam.exceptions.AzcamError("Did not receive header return status from remote image server")
        # retstat=int(reply[:1])
        # retstat=int(reply)
        retstat = 0

        # check header return status codes (updated 14jul11)
        if retstat != 0:
            if retstat == 1:  # overwrite existing name wihtout flag
                raise azcam.exceptions.AzcamError(
                    "Remote image server could not create image filename"
                )
            elif retstat == 2:  # not enough space
                raise azcam.exceptions.AzcamError(
                    "Remote image server does not have enough disk space"
                )
            elif retstat == 3:  #
                raise azcam.exceptions.AzcamError(
                    "Remote image server reports folder does not exist"
                )
            else:
                raise azcam.exceptions.AzcamError(
                    "Unknown error from remote image server"
                )

        # send file data, new 10Sep13
        numchunks = int(lSize / ImageSendBufferSize)
        if lSize - (numchunks * ImageSendBufferSize) != 0:
            remainder = lSize - (numchunks * ImageSendBufferSize)
        else:
            remainder = 0

        if numchunks == 0:
            size = remainder
        else:
            size = ImageSendBufferSize

        end = 0
        for _ in range(numchunks):
            start = end
            end = start + size
            try:
                # dataserver_socket.send(str.encode(buff[start:end]))
                dataserver_socket.send(buff[start:end])
            except Exception as message:
                raise azcam.exceptions.AzcamError(
                    f"Did not send image file data to remote image server: {message}"
                )
        if remainder > 0:
            # dataserver_socket.send(str.encode(buff[end:]))
            dataserver_socket.send(buff[end:])

        """
        # send file data
        try:
            reply=dataserver_socket.send(str.encode(buff))
        except Exception as message:
            raise azcam.exceptions.AzcamError("Did not send image file data to remote image server: %s" % message)

        if reply!=len(buff):
            raise azcam.exceptions.AzcamError("Did not send entire image file data to remote image server")
        """

        # get 16 char ASCII final return status from image server
        """
        try:
            reply = dataserver_socket.recv(16)
        except:
            raise azcam.exceptions.AzcamError("Did not receive return status from remote image server")

        if len(reply) != 2:
            raise azcam.exceptions.AzcamError("Did not receive entire return status from remote image server")
        retstat=int(reply[:1])

        # check final return status error codes
        if retstat != 0:
            raise azcam.exceptions.AzcamError("Bad final return status from remote image server")
        """

        # close socket
        time.sleep(1)  # 3
        dataserver_socket.close()

        return



"""
Used to bring commands into the current namespace.

Usage:  `from azcam.cli import *`

`azcam` as well as items in `db.cli` are 
added to __all__ here for import into the CLI namespace.
"""

import azcam

# main database object
db = azcam.db

# directly put tools in namespace to be imported with *
try:
    for name in azcam.db.cli:
        globals()[name] = azcam.db.cli[name]

    __all__ = [x for x in azcam.db.cli]
except Exception:
    pass

# __all__.append("azcam")
__all__.append("db")



"""
Contains the TempConArc class.
"""

import math

import azcam
import azcam.exceptions
from azcam.tools.tempcon import TempCon


class TempConArc(TempCon):
    """
    Defines the ARC temperature control class.
    This is used for Gen1, Gen2, and Gen3 ARC controllers.
    """

    def __init__(self, tool_id="tempcon", description=None):
        super().__init__(tool_id, description)

        self.num_temp_reads = 5

        self.temperature_ids = [0, 1]  # camtemp, dewtemp

        return

    def define_keywords(self):
        """
        Defines and resets tempcon keywords.
        """

        self.set_keyword("CAMTEMP", 0.0, "Camera temperature", "float")
        self.set_keyword("DEWTEMP", 0.0, "Dewar temperature", "float")
        self.set_keyword("TEMPUNIT", "C", "Temperature units", "str")

        return

    def set_control_temperature(self, temperature=None, temperature_id=-1):
        """
        Set controller/detector control temperature.
        Ignored if no utlity board is installed.
        Temperature is temperature to set in Celsius.
        """

        if not self.is_enabled:
            azcam.exceptions.warning(f"{self.description} is not enabled")
            return

        TEMPSET = 0x01C

        # ignore if no utlity board
        if not azcam.db.tools["controller"].utility_board_installed:
            return

        # ignore if controller is not reset
        if not azcam.db.tools["controller"].is_reset:
            return

        if temperature is None:
            temperature = self.control_temperature
        else:
            self.control_temperature = float(temperature)

        counts = self.convert_temp_to_counts(2, self.control_temperature)
        azcam.db.tools["controller"].write_memory(
            "Y", azcam.db.tools["controller"].UTILITYBOARD, TEMPSET, int(counts)
        )

        return

    def get_temperature(self, temperature_id=0):
        """
        Read a utlity board temperature.
        TemperaureID's are:
        0 => CAMTEMP
        1 => DEWTEMP
        2 => DIODETEMP
        """

        if not self.is_enabled:
            # azcam.exceptions.warning("Tempcon not enabled")
            return -999.9

        if not self.is_initialized:
            # azcam.exceptions.warning("Tempcon not initialized")
            return -999.9

        if not azcam.db.tools["controller"].utility_board_installed:
            return self.bad_temp_value

        if not azcam.db.tools["controller"].is_reset:
            return self.bad_temp_value

        # define DSP address
        if temperature_id == 0:
            Address = 13
        elif temperature_id == 1:
            Address = 14
        elif temperature_id == 2:
            Address = 12
        else:
            raise azcam.exceptions.AzcamError("bad temperature_id in get_temperature")

        temperature_id = int(temperature_id)

        # Don't read hardware while exposure is in progess
        flag = azcam.db.tools["exposure"].exposure_flag
        if flag != azcam.db.tools["exposure"].exposureflags["NONE"]:
            return self.last_temps[temperature_id]

        # read temperature
        cmd = "RDM"
        avecount = 0
        try:
            for _ in range(self.num_temp_reads):
                reply = azcam.db.tools["controller"].board_command(
                    cmd, azcam.db.tools["controller"].UTILITYBOARD, 0x400000 | Address
                )  # Y space
                counts = int(reply)
                avecount += counts
        except ValueError:
            raise azcam.exceptions.AzcamError("could not read temperature")
        counts = avecount / self.num_temp_reads

        # convert from counts to Celsius
        temp = self.convert_counts_to_temp(
            self.temperature_cals[temperature_id], counts
        )

        temp = self.apply_corrections(temp, temperature_id)

        # make nice float
        temp = float(int(temp * 1000.0) / 1000.0)

        # use some limits
        if temp > 100.0 or temp < -300.0:
            temp = -999.9

        # save temp
        self.last_temps[temperature_id] = temp

        return temp

    def convert_counts_to_temp(self, calflag: int, counts: int) -> float:
        """
        Convert counts (DN) to degrees Celsius.
        Uses Chebyechev polynomial.

        :param calflag: calibration curve to use
        :param counts: value to convert

          * 0 => convert DT670  counts to degrees C
          * 1 => convert AD590  counts to degrees C
          * 2 => convert 1N4148 counts to degrees C
          * 3 => convert 1N914  counts to degrees C
        """

        MAXBOUND = 11
        B = [
            82.017868,
            -59.064244,
            -1.356615,
            1.055396,
            0.837341,
            0.431875,
            0.440840,
            -0.061588,
            0.209414,
            -0.120882,
            0.055734,
            -0.035974,
        ]
        C = [
            306.592351,
            -205.393808,
            -4.695680,
            -2.031603,
            -0.071792,
            -0.437682,
            0.176352,
            -0.182516,
            0.064687,
            -0.027019,
            0.010019,
        ]
        A = MAXBOUND * [0]
        MAXDIODE = 19
        DC = [
            2600,
            2728,
            2823,
            2966,
            3003,
            3009,
            3070,
            3107,
            3118,
            3139,
            3166,
            3183,
            3217,
            3256,
            3269,
            3329,
            3357,
            3379,
            3417,
        ]  # diode count array
        DT = [
            311,
            277,
            250,
            210,
            200,
            190,
            180,
            170,
            160,
            155,
            150,
            147,
            137,
            127,
            117,
            107,
            97,
            87,
            77,
        ]  # diode temp array

        NOAO = [471.3507, -581.5675, 558.4024, -786.8676, 628.3720, -219.7995]

        VMIN = 0.07000
        VMAX = 1.122751
        VOLTMAX = 3.0
        VOLTMIN = -3.0
        COUNTMAX = 4095
        COUNTMIN = 0
        GAIN = 2.0  # this is the gain before ADC

        if calflag == 0:  # DT670 counts to Celsius
            # convert counts to voltage
            voltage = (
                ((VOLTMAX - VOLTMIN) / (COUNTMAX - COUNTMIN + 1))
                * (counts - 2048)
                / GAIN
            )

            if voltage > VMAX:
                temp = -999.9
                return temp
            elif voltage < VMIN:
                temp = +999.9
                return temp

            if voltage >= 0.986974:
                for i in range(MAXBOUND):  # 24.5 to 100.0 K
                    A[i] = B[i]
                    zl = 0.909416
                    zu = 1.122751
            else:
                for i in range(MAXBOUND):  # 100 to 500k
                    A[i] = C[i]
                    zl = 0.07000
                    zu = 0.99799

            x = ((voltage - zl) - (zu - voltage)) / (zu - zl)
            temp = 0.0
            for i in range(MAXBOUND):
                temp += A[i] * math.cos(i * math.acos(x))
            temp -= 273.13  # K to C
            return temp

        elif calflag == 1:  # AD590 case counts to Celsius
            temp = 0.16012 * counts - 599.5
            return temp

        elif calflag == 2:  # 1N4148 diode counts to Celsius
            if counts < DC[0]:
                temp = +999.9
                return temp
            elif counts > DC[MAXDIODE - 1]:
                temp = -999.9
                return temp

            for i in range(MAXDIODE):
                if counts < DC[i]:
                    inc = ((DT[i - 1] - DT[i]) / (DC[i - 1] - DC[i])) * (counts - DC[i])
                    temp = DT[i] + inc - 273.13  # simple for now
                    break
            return temp

        elif calflag == 3:  # 1N914 diode voltage to Celsius
            # convert counts to voltage
            voltage = (
                ((VOLTMAX - VOLTMIN) / (COUNTMAX - COUNTMIN + 1))
                * (counts - 2048)
                / GAIN
            )

            temp = (
                NOAO[0]
                + voltage * NOAO[1]
                + pow(voltage, 2) * NOAO[2]
                + pow(voltage, 3) * NOAO[3]
                + pow(voltage, 4) * NOAO[4]
                + pow(voltage, 5) * NOAO[5]
                - 273.13
            )
            return temp

        else:
            raise azcam.exceptions.AzcamError("ConvertCountsToTemp", "invalid calflag")

    def convert_temp_to_counts(self, calflag: int, temperature: float) -> float:
        """
        Convert degrees Celsius to counts.

        :param calflag: calibration flag to use
        :param temperature: temperature in C to convert
        """

        NOAOINV = [6.00e-09, -3.50e-06, -0.0021, 1.1616]  # inverse of NOAO

        if calflag == 2:  # 1N914 diode case Celsius to counts
            VOLTMAX = 3.0
            VOLTMIN = -3.0
            COUNTMAX = 4095
            COUNTMIN = 0
            GAIN = 2.0  # this is the gain before ADC

            inp = temperature + 273.13  # C to K
            voltage = (
                NOAOINV[3]
                + inp * NOAOINV[2]
                + pow(inp, 2) * NOAOINV[1]
                + pow(inp, 3) * NOAOINV[0]
            )
            counts = ((COUNTMAX - COUNTMIN + 1) / (VOLTMAX - VOLTMIN)) * (
                voltage * GAIN
            ) + 2048

        else:
            raise azcam.exceptions.AzcamError("convert_temp_to_counts invalid calflag")

        return counts



"""
Contains the Header and System classes.
"""

import os
import typing

import azcam
import azcam.utils
import azcam.exceptions


class Header(object):
    """
    Defines the Header class which is used to create and manipulate headers contained
    in many azcam tools.

    Attributes:
        self.title (str): header title dictionary {index:title_line}
        self.keywords (dict): header keywords dictionary {keyword:keywordstring}
        self.values (dict):  header values dictionary {keyword:value}
        self.comments (dict):  header comments dictionary {keyword:comment}
        self.typestrings (dict):  header typestrings dictionary {keyword:typestring)
        self.items (list):  list of header objects
    """

    def __init__(self, title: str = "", template: str | None = None):
        """
        Create instance.

        Args:
            title: title of header to be used in image header
            template: filename of template file with static header info
        """

        self.title = {}  # header title dictionary {index:title_line}
        self.keywords = {}  # header keywords dictionary {keyword:keywordstring}
        self.values = {}  # header values dictionary {keyword:value}
        self.comments = {}  # header comments dictionary {keyword:comment}
        self.typestrings = {}  # header typestrings dictionary {keyword:typestring)
        self.items = []  # list of header objects

        self.filename = ""  # image header filename

        if title != "":
            self.set_title(title)

        if template is not None:
            azcam.db.tools["exposure"].imageheaderfile = template
            self.read_file(template)

    def set_header(self, object_name: str, order: int = -1):
        """
        Sets object_name in the global header dictionary.
        order defines how headers are written in image files.

        Args:
            object_name: name of header object
            order: order to be shown in image header (-1 next)
        """

        if object_name not in list(
            azcam.db.headers.keys()
        ):  # do not change order if already exists
            if order == -1:
                azcam.db.headerorder.append(object_name)
            else:
                azcam.db.headerorder.insert(order, object_name)

        azcam.db.headers[object_name] = self  # may replace an existing object

        return

    def set_title(self, title):
        """
        Set the title of the header.
        """

        # special case
        title = "AzCam Focal plane" if title == "Focalplane" else title
        self.title[0] = (
            "=================================================================="
        )
        self.title[1] = f"{title}"
        self.title[2] = (
            "=================================================================="
        )

        return

    def copy_all_items(self, ptr_header):
        """
        Copy all items from the specified header object to the current image header.
        """

        self.items.append(ptr_header)

        return

    def delete_all_items(self):
        """
        Delete all header items
        """

        self.items = []

        return

    def get_keywords(self):
        """
        Return a list of all keyword names.
        Returns:
            keywords: list of all keywords
        """

        if len(self.keywords) > 0:
            klist = list(self.keywords.keys())
            klist.sort()
        else:
            klist = []

        return klist

    def set_keyword(
        self,
        keyword: str,
        value: typing.Any,
        comment: str | None = None,
        typestring: str | None = None,
    ):
        """
        Set a keyword value, comment, and type.
        Args:
            keyword: keyword
            value: value of keyword
            comment: comment string
            typestring: one of 'str', 'int', or 'float'
        """

        if keyword not in self.keywords:
            self.keywords[keyword] = keyword
            self.values[keyword] = None

        if isinstance(value, str):
            value = value.strip()

        if typestring is None:
            if self.typestrings.get(keyword) is None:
                self.typestrings[keyword] = "str"
        else:
            self.typestrings[keyword] = typestring

        if value is not None:
            if self.typestrings[keyword] == "str":
                value = str(value)
            elif self.typestrings[keyword] == "int":
                value = int(value)
            elif self.typestrings[keyword] == "float":
                value = float(value)
        self.values[keyword] = value

        if comment is None or comment == "none":
            if self.comments.get(keyword):  # use previous comment
                pass
            else:
                self.comments[keyword] = ""
        else:
            self.comments[keyword] = comment

        return

    def set_keyword_string(self, keystring):
        """
        Set keyword data from a single string.
        """

        keystring = keystring.strip()
        if len(keystring) == 0:
            return

        # allow for quotes
        tokens = azcam.utils.parse(keystring)

        keyword = tokens[0]

        # find comment after '/'
        comment = ""
        value = ""
        for i, t in enumerate(tokens):
            if t == "/":
                comment = " ".join(tokens[i + 1 :])
                value = " ".join(tokens[1:i])

        if comment == "":
            value = " ".join(tokens[1:])

        # try to get type of value
        t = "str"
        if isinstance(value, int):
            value = int(value)
            t = "int"
        elif isinstance(value, float):
            value = float(value)
            t = "float"

        self.set_keyword(keyword, value, comment, t)

        return

    def get_keyword(self, keyword: str) -> list:
        """
        Return a keyword value, its comment string, and type.
        Comment always returned in double quotes, even if empty.
        Args:
            keyword (str): name of keyword
        Returns:
            list of [keyword, comment, type]
        """

        value = self.values[keyword]
        comment = self.comments[keyword]
        if comment is None:
            comment = ""
        t = self.typestrings[keyword]

        return [value, comment, t]

    def delete_keyword(self, keyword: str):
        """
        Delete a keyword.
        """

        try:
            del self.keywords[keyword]
        except KeyError:
            pass
        try:
            del self.values[keyword]
        except KeyError:
            pass
        try:
            del self.comments[keyword]
        except KeyError:
            pass

        return

    def delete_all_keywords(self):
        """
        Delete all keywords.
        """

        keys = self.get_keywords()
        for key in keys:
            self.delete_keyword(key)  # delete in header

        return

    def get_string(self):
        """
        Returns the entire header as a single formatted string.
        """

        lines = ""

        header = self.get_header()
        for telem in header:
            line = telem[0] + " " + str(telem[1]) + " " + str(telem[2]) + "\n"
            lines += line

        return lines

    def convert_type(self, value, pythontype):
        """
        Convert a value to a specific type.
        """

        if pythontype == "str" or pythontype not in ["int", "float"]:
            typestring = "str"
            value = str(value)
        elif pythontype == "int":
            typestring = "int"
            value = int(value)
        else:
            typestring = "float"
            value = float(value)

        return value, typestring

    def update(self):
        """
        Update header.
        """

        return

    def get_header(self) -> list:
        """
        Returns the header list.
        Returns:
            list of header lines: [Header[]]: Each element contains (keyword,value,comment,type).
        """

        # get the header
        header = []
        reply = self.get_keywords()
        if reply == []:
            return header

        for key in reply:
            reply1 = self.get_keyword(key)
            list1 = [key, reply1[0], reply1[1], reply1[2]]
            header.append(list1)

        return header

    def read_file(self, filename=""):
        """
        Read a header file and import the data into the header.
        """

        # if no file specified, just skip
        if filename == "":
            return

        if not os.path.exists(filename):
            azcam.exceptions.AzcamError(f"Header file not found: {filename}")

        with open(filename, "r") as f1:
            for line in f1.readlines():
                line = line.rstrip()
                line = line.lstrip()
                if line.startswith("#"):
                    continue
                if len(line) == 0:
                    continue
                tokens = line.split(" ", 1)
                keyword = tokens[0]

                if keyword == "COMMENT" and len(tokens) == 1:
                    value = ""
                    comment = ""
                else:
                    nslash = tokens[1].find("/")
                    if nslash == -1:
                        value = tokens[1].strip()
                        comment = ""
                    else:
                        comment = tokens[1][nslash + 1 :].strip()
                        value = tokens[1][:nslash].strip()

                typ, val = azcam.utils.get_datatype(value)
                self.set_keyword(keyword, val, comment, typ)

        self.filename = filename

        return


class ObjectHeaderMethods(object):
    """
    Header methods for main objects.
    These are called like "controller.get_keyword()".
    """

    def __init__(self) -> None:
        pass

    def define_keywords(self):
        """
        Defines and resets keywords.
        """

        return

    def update_header(self):
        """
        Update the header, reading current data.
        Deletes all keywords if the object is not enabled.
        """

        if not self.is_enabled:
            self.header.delete_all_keywords()
            return

        if not self.is_initialized:
            self.initialize()

        self.define_keywords()

        self.read_header()

        return

    def read_header(self) -> list:
        """
        Reads and returns current header data.
        Returns:
            list of header lines: [Header[]]: Each element contains (keyword,value,comment,type).
        """

        # Example: Header[2][1] is the value of keyword 2 and Header[2][3] is its type.

        # get the header
        header = []
        reply = self.header.get_keywords()
        if reply == []:
            return

        for key in reply:
            reply1 = self.get_keyword(
                key
            )  # this calls object's get_keyword to get updated values
            list1 = [key, reply1[0], reply1[1], reply1[2]]
            header.append(list1)

        return header

    def get_keyword(self, keyword: str) -> list:
        """
        Return a keyword value, its comment string, and type.
        Comment always returned in double quotes, even if empty.
        Args:
            keyword: name of keyword
        Returns:
            list of [keyword, comment, type]

        """

        return self.header.get_keyword(keyword)

    def get_keywords(self) -> list:
        """
        Return a list of all keyword names.
        Returns:
            keywords: list of all keywords
        """

        return self.header.get_keywords()

    def set_keyword(
        self,
        keyword: str,
        value: typing.Any,
        comment: str = "none",
        typestring: str = "none",
    ):
        """
        Set a keyword value, comment, and type.
        Args:
            keyword: keyword
            value: value of keyword
            comment: comment string
            typestring: one of 'str', 'int', or 'float'
        """

        # if typestring == "none":
        #     typestring = None
        # if comment == "none":
        #     comment = None

        self.header.set_keyword(keyword, value, comment, typestring)

        return

    def delete_keyword(self, keyword: str):
        """
        Delete a keyword.
        Args:
            keyword: keyword
        """

        self.header.delete_keyword(keyword)

        return


class System(ObjectHeaderMethods):
    """
    System class.
    """

    def __init__(self, system_name, template_file=None):
        self.header = Header(system_name, template_file)
        self.header.set_header("system", 0)

        self.system = self

        return



"""
Contains the base TempCon class.
"""

from typing import List

import azcam
import azcam.exceptions
from azcam.header import Header, ObjectHeaderMethods
from azcam.tools.tools import Tools
from typing import Union, List, Optional


class TempCon(Tools, ObjectHeaderMethods):
    """
    The base temperature control tool.
    Usually implemented as the "tempcon" tool.
    """

    def __init__(self, tool_id: str = "tempcon", description: str = None):
        """
        Creates the tool.

        Args:
            tool_id: Name of tool
            description: Description of tool, defaults to tool_id.
        """

        Tools.__init__(self, tool_id, description)

        self.control_temperature = -999.0
        """control temperature in Celsius"""
        self.control_temperature_id = 0
        """sensor used for temperature control"""

        # system temperatures
        self.temperature_ids = [0]

        # True to correct temperature offset and scale for each temperature
        # Scaling is **new = old * scale + offset**
        self.temperature_correction = 0
        # temperature offset corrections for each temperature
        self.temperature_offsets = [0.0]
        # temperature scale corrections
        self.temperature_scales = [1.0]
        # calibration flags for each temperature
        self.temperature_cals = [3]

        # value returned when temperature read is bad
        self.bad_temp_value = -999.9

        self.last_temps = 3 * [self.bad_temp_value]  # last readings for during exposure

        # create the temp control Header object
        self.header = Header("Temperature")
        self.header.set_header("tempcon", 4)

        # add keywords
        self.define_keywords()

        azcam.db.tools_init["tempcon"] = self
        azcam.db.tools_reset["tempcon"] = self

        return

    def initialize(self):
        """
        Initialize tool.
        """

        self.is_initialized = 1

        return

    def reset(self) -> None:
        """
        Reset tempcon tool.
        """

        if not self.is_enabled:
            return

        if not self.is_initialized:
            self.initialize()

        self.set_control_temperature()

        return

    # ***************************************************************************
    # exposure
    # ***************************************************************************
    def exposure_start(self) -> None:
        """
        Custom commands before exposure starts.
        """

        return

    def exposure_finish(self) -> None:
        """
        Custom commands after exposure finishes.
        """

        return

    # ***************************************************************************
    # temperatures
    # ***************************************************************************
    def set_control_temperature(
        self, temperature: float = None, temperature_id: int = -1
    ) -> None:
        """
        Set the control temperature (set point).
        Args:
            temperature: control temperature in Celsius. If not specified, use saved value
            temperature_id: temperature sensor identifier
        """

        if temperature is None:
            temperature = self.control_temperature
        else:
            self.control_temperature = temperature

        return

    def get_control_temperature(self, temperature_id: int = 0) -> float:
        """
        Get the control temperature (set point).
        Args:
            temperature_id: temperature sensor identifier
        Returns:
            control_temperature: control temperature
        """

        return self.control_temperature

    def get_temperatures(self) -> List[float]:
        """
        Return all system temperatures.
        Returns:
            temperatures: list of temperatures read
        """

        temps = []
        for temperature_id in self.temperature_ids:
            p = self.get_temperature(temperature_id)
            temps.append(p)

        return temps

    def get_temperature(self, temperature_id: int = 0) -> float:
        """
        Returns a system temperature.
        Args:
            temperature_id: temperature sensor identifier
        Returns:
            temperature: temperature read
        """

        return self.bad_temp_value

    # ***************************************************************************
    # calibrations
    # ***************************************************************************
    def set_calibrations(self, cals: List[int]) -> None:
        """
        Set calibration curves for temperature sensors.
        The values of these flags are from the list below which define the calibration curves to use
        for each sensor's temperature conversion.

          * 0 => DT670  diode
          * 1 => AD590  sensor
          * 2 => 1N4148 diode
          * 3 => 1N914  diode
        Args:
            cals: list of flags defining each temperature sensor type
        """

        self.temperature_cals = []
        for i, cal in enumerate(cals):
            self.temperature_cals.append(cal)

        return

    def set_corrections(
        self,
        temperature_offsets: List[float] = None,
        temperature_scales: List[float] = None,
    ) -> None:
        """
        Set temperature correction values and enable correction.
        If both parameters are None then correction is disabled.
        Args:
            temperature_offsets: list of offsets for each temperature
            temperature_scales: list of scale factors for each temperature
        """

        # set and save values
        if temperature_offsets is None and temperature_scales is None:
            self.temperature_correction = 0
        else:
            self.temperature_correction = 1
        if temperature_offsets is not None:
            self.temperature_offsets = temperature_offsets
        if temperature_scales is not None:
            self.temperature_scales = temperature_scales

        return

    def apply_corrections(self, temperature: float, temperature_id: int = 0) -> float:
        """
        Correct the temperatures for offset and scale is temperature correction is enabled.
        Args:
            temperature: temperatures to be corrected
            temperature_id: temperature ID number
        Returns:
            corrected_temperature: temperature after correction has been appied
        """

        if self.temperature_correction:
            return (
                temperature * self.temperature_scales[temperature_id]
                + self.temperature_offsets[temperature_id]
            )
        else:
            return temperature

    # ***************************************************************************
    # keywords
    # ***************************************************************************
    def define_keywords(self):
        """
        Defines and resets tempcon keywords.
        """

        return

    def get_keyword(self, keyword: str) -> List:
        """
        Read a temperature keyword value and returns it as [value, comment, type string]
        Args:
            keyword: name of keyword
        Returns:
            list of [keyword, comment, type]
        """

        reply = self.get_temperatures()

        if keyword == "CAMTEMP":
            temp = reply[0]
        elif keyword == "DEWTEMP":
            temp = reply[1]
        elif keyword in self.get_keywords():
            value = self.header.values[keyword]
            temp = value
        else:
            raise azcam.exceptions.AzcamError(f"invalid keyword: {keyword}")

        # store temperature values in header
        if keyword == "CAMTEMP":
            temp = float(f"{temp:.03f}")
            self.header.set_keyword("CAMTEMP", temp, "Camera temperature in C", "float")
            self.header.set_keyword(
                "CCDTEMP1", temp, "Camera temperature in C", "float"
            )
        elif keyword == "DEWTEMP":
            temp = float(f"{temp:.03f}")
            self.header.set_keyword("DEWTEMP", temp, "Dewar temperature in C", "float")
            self.header.set_keyword("CCDTEMP2", temp, "Dewar temperature in C", "float")
        t = self.header.typestrings[keyword]

        return [temp, self.header.comments[keyword], t]



"""
AzCam is a software framework for the acquisition and analysis of image data
from scientific imaging systems as well as the control of instrumentation.
"""

import typing
from typing import List, Dict
from importlib import metadata

__version__ = metadata.version(__package__)
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())

from azcam.logger import AzCamLogger

from azcam.database import AzcamDatabase

# logger object
logger: AzCamLogger = AzCamLogger()

# initially azcam.log() is print() but will usually be overwritten
log: typing.Callable = print

# initial database but will ususally be overwritten by server or console
db: AzcamDatabase = AzcamDatabase()

mode = "server"
"""azcam mode, usually server or console"""

# cleanup namespace
del metadata
del typing
del AzCamLogger
del AzcamDatabase
# del database



import os


class Filename(object):
    """
    Filename class used to manipulate image filenames.
    Used by the exposure tool.
    """

    def __init__(self, parent=None):
        """
        Creates a filename instance.

        param object parent: reference to parent image object
        """

        # self.parent = parent  # calling object reference
        self.parent = self  # calling object reference

        self.filetype = 0  # from image object
        # True to overwrite image files of the same name
        self.overwrite = 0
        # current image file folder
        self.folder = ""
        # current image file root name
        self.root = "a."
        # current image file sequence number
        self.sequence_number = 1
        # True to increment file sequence number after each exposure
        self.auto_increment_sequence_number = 0
        # True to include file sequence number in each image name
        self.include_sequence_number = 0
        # True to automatically name image based on imagetype
        self.autoname = 0
        # True when current file is a test image
        self.test_image = 1

        if parent is not None:
            self.filetype = parent.filetype

    def get_filename(self):
        """
        Return current filename as a single string.
        This is the filename for the next exposure to be taken.

        Always uses forward slashes for folder delimiter.
        This name is usually the next image to be produced.
        """

        self.filetype = self.parent.filetype

        folder = self.folder.replace("\\", "/")
        if not folder.endswith("/"):
            folder = folder + "/"

        extension = self.get_extname(self.filetype)

        if self.test_image:
            filename = folder.replace("\\", "/") + "test" + "." + extension
        elif self.include_sequence_number:
            if self.autoname:
                filename = (
                    folder
                    + self.root
                    + "."
                    + self.parent.image_type.upper()
                    + "."
                    + "%04d" % self.sequence_number
                    + "."
                    + extension
                )
            else:
                filename = (
                    folder + self.root + "%04d" % self.sequence_number + "." + extension
                )
        else:
            if self.autoname:
                filename = (
                    folder
                    + self.root
                    + "."
                    + self.parent.image_type.upper()
                    + "."
                    + extension
                )
            else:
                filename = folder + self.root + "." + extension

        filename = filename.replace(
            "..", "."
        )  # clean up as could have two periods together

        return filename

    def set_filename(self, filename: str):
        """
        Set the filename components based on a simple filename.

        Args:
            filename: filename to be set
        """

        self.filetype = self.parent.filetype

        filename = os.path.normpath(filename)
        self.folder = os.path.dirname(filename)
        if self.folder == "":
            self.folder = "./"

        f = os.path.basename(filename)

        # if f.endswith(".fits"):
        #     f = f.replace(".fits", "")
        #     self.filetype = 0

        f = f.split(".")  # root is just first part
        self.root = f[0]

        return

    def increment_filenumber(self):
        """
        Increment the filename sequence number if AutoIncrementSequenceNumber is True and not a test image.
        """

        if self.auto_increment_sequence_number and not self.test_image:
            self.sequence_number += 1

        return

    def get_extname(self, filetype):
        """
        Return the file extension string for a file type.
        """

        if filetype in [0, 1, 6]:
            return "fits"
        elif filetype == 2:
            return "bin"
        elif filetype == 3:
            return "tif"
        elif filetype == 4:
            return "jpg"
        elif filetype == 5:
            return "gif"
        else:
            return ""



from dash import html, dcc, callback, Input, Output
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam


def create_button_group():
    """
    Create button group for exposure control.
    """

    button_group = dbc.ButtonGroup(
        [
            dbc.Button("Expose", id="expose_btn", className="m-1", n_clicks=0),
            dbc.Button("Sequence", id="sequence_btn", className="m-1", n_clicks=0),
            dbc.Button("Reset", id="reset_btn", className="m-1", n_clicks=0),
            dbc.Button("Abort", id="abort_btn", className="m-1", n_clicks=0),
            dbc.Button("Save Pars", id="savepars_btn", className="m-1", n_clicks=0),
        ],
        vertical=True,
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("expose_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_expose(n):
        try:
            azcam.db.tools["exposure"].expose()
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("sequence_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_sequence(n):
        try:
            azcam.db.tools["exposure"].sequence()
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("reset_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_reset(n):
        try:
            azcam.db.tools["exposure"].reset()
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("abort_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_abort(n):
        try:
            azcam.db.tools["exposure"].abort()
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("savepars_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_savepars(n):
        try:
            azcam.db.parameters.save_pars()
        except Exception as e:
            print(e)
        return

    return button_group


def exposure_card():
    """
    Create exposure card for exposure control.
    """

    options = [
        {"label": x, "value": x.lower()} for x in azcam.db.tools["exposure"].image_types
    ]

    image_type_dropdown = html.Div(
        [
            dbc.Label("Image Type"),
            dcc.Dropdown(
                options,
                id="image_type_dropdown",
                value="zero",
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("image_type_dropdown", "value"),
        prevent_initial_call=True,
    )
    def image_type_callback(value):
        azcam.db.tools["exposure"].set_image_type(value)
        return ""

    exposure_time_input = html.Div(
        [
            dbc.Label("Exp. time [sec]"),
            daq.NumericInput(id="et", value=0.0, size=100),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("et", "value"),
        prevent_initial_call=True,
    )
    def exposure_time_callback(value):
        try:
            azcam.db.tools["exposure"].set_exposuretime(value)
        except Exception:
            pass  # controller may not be reset
        return ""

    image_title_input = html.Div(
        [
            dbc.Label("Image Title"),
            dbc.Input(
                id="title_input",
                placeholder="Enter image title...",
                type="text",
                style={"display": "inline-block"},
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("title_input", "value"),
        prevent_initial_call=True,
    )
    def image_title_callback(value):
        azcam.db.tools["exposure"].set_image_title(value)
        return value

    test_image_input = html.Div(
        [
            dbc.Label("Test Image"),
            dbc.Checkbox(id="test_image", value=True),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("test_image", "value"),
        prevent_initial_call=True,
    )
    def image_test_callback(value):
        azcam.db.tools["exposure"].set_test_image(value)
        return value

    ###########################################################################
    # sequence pars
    ###########################################################################
    num_seq_images_input = html.Div(
        [
            daq.NumericInput(id="seq_num", label="Number seq. images", size=100),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("seq_num", "value"),
        prevent_initial_call=True,
    )
    def seq_total_callback(value):
        azcam.db.tools["exposure"].exposure_sequence_total = int(value)
        return value

    sequence_delay_input = html.Div(
        [
            daq.NumericInput(
                id="seq_delay",
                label="Seq. delay [sec]",
                size=100,
                style={"margin-top": "2"},
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("seq_delay", "value"),
        prevent_initial_call=True,
    )
    def seq_delay_callback(value):
        azcam.db.tools["exposure"].exposure_sequence_delay = float(value)
        return value

    button_group = create_button_group()

    exposure_card = dbc.Card(
        [
            dbc.CardHeader("Exposure Control", style={"font-weight": "bold"}),
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            dbc.Col(button_group, width=2),
                            dbc.Col(
                                [
                                    dbc.Row(
                                        [
                                            dbc.Col(image_type_dropdown),
                                            dbc.Col(test_image_input),
                                            dbc.Col(exposure_time_input),
                                        ]
                                    ),
                                    dbc.Row(
                                        [image_title_input],
                                    ),
                                    dbc.Row(
                                        [
                                            dbc.Col(num_seq_images_input, width=3),
                                            dbc.Col(sequence_delay_input, width=6),
                                        ]
                                    ),
                                ]
                            ),
                        ]
                    ),
                ]
            ),
        ]
    )

    return exposure_card



import requests

if 1:
    data1 = {
        "tool": "parameters",
        "command": "get_par",
        "args": ["imagetest"],
        "kwargs": {},
    }

    data2 = {
        "tool": "parameters",
        "command": "set_par",
        "args": ["imagetest", "2222"],
        "kwargs": {},
    }

    data3 = {
        "tool": "parameters",
        "command": "set_par",
        "args": [],
        "kwargs": {"parameter": "imagetest", "value": 3333},
    }

    data4 = {
        "tool": "parameters",
        "command": "get_par",
        "args": ["imagetest"],
        "kwargs": {},
    }

    r = requests.post("http://localhost:2403/api", json=data1)
    print(r.status_code, r.json())

    r = requests.post("http://localhost:2403/api", json=data2)
    print(r.status_code, r.json())

    r = requests.post("http://localhost:2403/api", json=data3)
    print(r.status_code, r.json())

    r = requests.post("http://localhost:2403/api", json=data4)
    print(r.status_code, r.json())

    r = requests.get("http://localhost:2403/api/exposure/get_status")
    print(r.status_code, r.json())



"""
Parameter handling tool for azcamserver.
"""

import typing

import azcam
from azcam.parameters import Parameters
import azcam.utils


class ParametersServer(Parameters):
    """
    Main class for server parameters tool.
    This tool sets the default subdict to *azcamserver*.
    """

    def __init__(self):
        """
        Creates parameters tool, optionally setting default parameter dictionary name.
        """

        Parameters.__init__(self, "azcamserver")

    def _get_par_hook(self, parameter, subdict):
        """
        Return the value of a parameter for server special cases.
        """

        value = None

        if parameter == "wd":
            value = azcam.utils.curdir()
        elif parameter == "logdata":
            value = azcam.logger.get_logdata()
        elif parameter == "imagefilename":
            value = azcam.db.tools["exposure"].get_filename()
        elif parameter == "imagetitle":
            value = azcam.db.tools["exposure"].get_image_title()
        elif parameter == "exposuretime":
            value = azcam.db.tools["exposure"].get_exposuretime()
        elif parameter == "exposurecompleted":
            value = azcam.db.tools["exposure"].finished()
        elif parameter == "exposuretimeremaining":
            value = azcam.db.tools["exposure"].get_exposuretime_remaining()
        elif parameter == "pixelsremaining":
            value = azcam.db.tools["exposure"].get_pixels_remaining()
        elif parameter == "camtemp":
            value = azcam.db.tools["tempcon"].get_temperatures()[0]
        elif parameter == "dewtemp":
            value = azcam.db.tools["tempcon"].get_temperatures()[1]
        elif parameter == "temperatures":
            camtemp = azcam.db.tools["tempcon"].get_temperatures()[0]
            dewtemp = azcam.db.tools["tempcon"].get_temperatures()[1]
            value = [camtemp, dewtemp]
        elif parameter == "logcommands":
            value = azcam.db.cmdserver.logcommands
        else:
            raise AttributeError

        return value

    def _set_par_hook(self, parameter, value, subdict):
        """
        Sets the value of a parameter for server special cases.
        """

        # special cases
        if parameter == "imagefilename":
            value = azcam.utils.dequote(value)
            azcam.db.tools["exposure"].image.filename = value
        elif parameter == "imagetitle":
            if value is None or value == "" or value == "None":
                azcam.db.tools["exposure"].set_image_title("")
            else:
                value = azcam.utils.dequote(value)
                azcam.db.tools["exposure"].set_image_title(value)
        elif parameter == "autotitle":
            azcam.db.tools["exposure"].set_auto_title(int(value))
        elif parameter == "imagetype":
            value = azcam.utils.dequote(value)
            azcam.db.tools["exposure"].image_type = value
            azcam.db.tools["exposure"].set_image_title()
        elif parameter == "exposuretime":
            azcam.db.tools["exposure"].set_exposuretime(value)
        elif parameter == "logcommands":
            azcam.db.cmdserver.logcommands = int(value)
        elif parameter == "wd":
            azcam.utils.curdir(value)
        else:
            raise AttributeError

        return None

    # TODO - below is for compatibility with azcamtool only - to be removed

    def set_script_par(self, attribute, value, subdict) -> None:
        azcam.db.parameters.par_dict[subdict][attribute] = value
        return

    def get_script_par(self, subdict, attribute) -> typing.Any:
        reply = azcam.db.parameters.par_dict[subdict][attribute]
        return reply



import importlib
import importlib.util
import os

import azcam
import azcam.exceptions


def loadscripts(packages: list) -> None:
    """
    Load scripts into azcam.db.scripts.
    """

    for package in packages:
        pyfiles = []
        spec = importlib.util.find_spec(package)
        if spec is not None:
            folder = spec.submodule_search_locations[0]
        else:
            return

        # bring all .py modules with same function name into namespace
        _, _, filenames = next(os.walk(folder))
        for file1 in filenames:
            if file1.endswith(".py"):
                pyfiles.append(file1[:-3])
        if "__init__" in pyfiles:
            pyfiles.remove("__init__")

        for pfile in pyfiles:
            try:
                mod = importlib.import_module(f"{package}.{pfile}")
                func = getattr(mod, pfile)
                # azcam.db.scripts[pfile] = func
                azcam.db.cli[pfile] = func
            except Exception as e:
                azcam.log(e)
                azcam.exceptions.warning(f"Could not import script {pfile}")

    return



"""
Parameter handling tool for azcam.

There is one main paramater dictionary with multiple subdicts. A default subdict can be specified.  
"""

import configparser
import os
import typing

import azcam
import azcam.utils
import azcam.exceptions


class Parameters(object):
    """
    Main class for parameters tool.
    """

    def __init__(self, default_dictname: str = None):
        """
        Creates parameters tool, optionally setting default parameter dictionary name.
        """

        self.par_file = None
        self.par_dict = {}

        self.default_pardict_name = default_dictname

        azcam.db.cli["parameters"] = self

    def read_parfile(self, parfilename: str = None) -> None:
        """
        Read a parameter file and create sub-dictionaries for saving parameters between sessions.

        Args:
            parfilename: Name of parameter file
        """

        if parfilename is None:
            parfilename = self.par_file
            if parfilename is None:
                azcam.exceptions.warning("Parameter file is not defined")
                return

        self.par_file = parfilename

        if not os.path.exists(parfilename):
            azcam.exceptions.warning(f"Parameter file not found: {parfilename}")
            return

        cp = configparser.ConfigParser()
        cp.read(parfilename)

        sections = cp.sections()

        # sectionname & value case sensitive, name is not
        for sectionname in sections:
            self.par_dict[sectionname] = {}

            for name, value in cp.items(sectionname):
                name = name.lower()
                self.par_dict[sectionname][name] = value

        return

    def write_parfile(self, parfilename: str | None = None) -> None:
        """
        Writes par_dict to the par file.
        Does not update any values.

        Args:
            parfilename (str, optional): Name of parameter file. Defaults to None.
                None indicates use self.par_file

        Raises:
            FileNotFoundError: if parfilename not found
        """
        if parfilename is None:
            parfilename = self.par_file
            if parfilename is None:
                raise FileNotFoundError("Parameter file is not defined")

        config = configparser.ConfigParser()

        for sectionname in self.par_dict:
            for par in self.par_dict[sectionname]:
                if self.par_dict[sectionname][par] is None:
                    self.par_dict[sectionname][par] = "None"
            config[sectionname] = self.par_dict[sectionname]

        # write parfile
        with open(parfilename, "w") as configfile:
            config.write(configfile)

        return

    def save_pars(self) -> None:
        """
        Writes the par_dict to the par_file using current values.
        """

        self.update_par_dict()
        self.write_parfile()

        return

    def update_pars(self) -> None:
        """
        Set current attributes from default par_dict values.
        """

        par_dictname = self.default_pardict_name

        par_dict = azcam.db.parameters.par_dict.get(par_dictname)
        if par_dict is None:
            return
        keys = par_dict.keys()
        if keys is None:
            return

        for parname in par_dict:
            value = par_dict[parname]
            if parname == "wd":
                azcam.db.wd = value
                azcam.utils.curdir(value)
            else:
                value = par_dict[parname]
                self.set_par(parname, value)

        return

    def update_par_dict(self) -> None:
        """
        Set par_dict values in default par_dict from current attributes.
        """

        par_dictname = self.default_pardict_name

        par_dict = azcam.db.parameters.par_dict.get(par_dictname)
        if par_dict is None:
            return
        keys = par_dict.keys()
        if keys is None:
            return

        for parname in par_dict:
            if parname == "wd":
                value = azcam.utils.curdir()
            else:
                value = self.get_par(parname)
            if value is None:
                value = "None"
            par_dict[parname] = value

        return

    def get_par(self, parameter: str, subdict=None) -> typing.Any:
        """
        Return the current attribute value of a parameter in the parameters dictionary.
        If subdict is not specified then the default subdict is used.

        Args:
            parameter: name of the parameter
            subdict: name of the subdict containing the parameter

        Returns:
            value: value of the parameter
        """

        parameter = parameter.lower()
        value = None

        if subdict is None:
            subdict = self.default_pardict_name

        # special cases hook
        try:
            value = self._get_par_hook(parameter, subdict)
            return value
        except NameError:  # OK if hook is not defined
            pass
        except AttributeError:  # OK if parameter not in hook
            pass

        # check if parameter is in par_table
        try:
            attribute = azcam.db.par_table[parameter]
            tokens = attribute.split(".")
            numtokens = len(tokens)

            # a tool and attribute is required
            if numtokens == 1:
                return None

            object1 = tokens[0]

            # object1 must be a tool or the database
            if object1 == "db":
                obj = azcam.db
            else:
                obj = azcam.db.tools[object1]
            for i in range(1, numtokens):
                try:
                    obj = getattr(obj, tokens[i])
                except AttributeError:
                    pass
            value = obj  # last time is value

        except KeyError:
            # check if value is known directly
            try:
                value = azcam.db.parameters.par_dict[subdict][parameter]
            except KeyError:
                azcam.exceptions.warning(
                    f"Parameter {parameter} not available for get_par"
                )
                return None

        return value

    def set_par(self, parameter: str, value: typing.Any = "None", subdict=None) -> None:
        """
        Set the value of a parameter in a parameters dictionary.
        Also attempts to set actual attribute value.

        Args:
            parameter: name of the parameter
            value: value of the parameter. Defaults to None.
            subdict: subdict in which to set paramater
        """

        parameter = azcam.utils.dequote(parameter)
        parameter = parameter.lower()

        if subdict is None:
            subdict = self.default_pardict_name

        # special cases hook
        try:
            self._set_par_hook(parameter, value, subdict)
            return None
        except NameError:  # OK if hook is not defined
            pass
        except AttributeError:  # OK if parameter not in hook
            pass

        # check if parameter is in par_table
        try:
            attribute = azcam.db.par_table[parameter]

            # object must be a tool
            tokens = attribute.split(".")
            numtokens = len(tokens)
            if numtokens < 2:
                azcam.log("%s not valid for parameter %s" % (attribute, parameter))
                return None

        except KeyError:
            _, value = azcam.utils.get_datatype(value)

            try:
                azcam.db.parameters.par_dict[subdict][parameter] = value
            except KeyError:
                if azcam.db.parameters.par_dict.get(subdict) is None:
                    azcam.db.parameters.par_dict[subdict] = {}
                azcam.db.parameters.par_dict[subdict][parameter] = value
            return None

        # first try to set value type
        _, value = azcam.utils.get_datatype(value)
        object1 = tokens[0]

        # run through tools
        try:
            obj = azcam.db.tools[object1]
            for i in range(1, numtokens - 1):
                obj = getattr(obj, tokens[i])
            # last time is actual object
            try:
                setattr(obj, tokens[-1], value)
            except AttributeError:
                pass
        except KeyError:
            pass
        except Exception:  # new
            pass

        return None

    def _get_par_hook(self, parameter, subdict):
        """
        Return the value of a parameter for special cases.
        """

        value = None

        return value

    def _set_par_hook(self, parameter, value, subdict):
        """
        Sets the value of a parameter for special cases.
        """

        return None



"""
azcam image I/O routines.
"""

import os
import time
import warnings

import numpy
from astropy.io import fits as pyfits


class ImageIO(object):
    """
    Methods for azcam image I/O.
    """

    def __init__(self):
        super().__init__()

    def _read_fits_file(self, filename):
        """
        Reads an image from a FITS or MEF disk file into the image structure.
        """

        self.assembled = 0
        self.is_valid = 0
        self.toggle = 0

        # clear focal plane header
        self.focalplane.header.delete_all_items()

        self.hdulist = pyfits.open(filename)
        if len(self.hdulist) == 2:
            NumExt = 0
            first_ext = 0
            last_ext = 1
        else:
            n = 0
            for i in range(len(self.hdulist)):
                if "XTENSION" in self.hdulist[i].header:
                    if self.hdulist[i].header["XTENSION"] == "IMAGE":
                        n += 1
            first_ext = 1
            last_ext = n + 1
            NumExt = n

        self.NumBinTables = 0  # temp
        self.num_extensions = NumExt

        # update number of amplifiers and number of extensions
        self.focalplane.numamps_image = max(1, NumExt) - self.NumBinTables
        cntExt = self.focalplane.numamps_image

        # update file type
        if self.num_extensions > 0:
            self.filetype = self.filetypes["MEF"]
        else:
            self.filetype = self.filetypes["FITS"]

        # get main header
        hdr = self.hdulist[0].header

        if self.num_extensions > 0:
            self.bitpix = self.hdulist[1].header["BITPIX"]
            try:
                self.bzero = self.hdulist[1].header["BZERO"]
                self.bscale = self.hdulist[1].header["BSCALE"]
            except KeyError:
                self.bzero = 0
                self.bscale = 0

            self.data_type = self.hdulist[1].data.dtype
            self.bitpix2 = self.hdulist[1].header["BITPIX"]
            self.save_data_format = self.bitpix
        else:
            self.bitpix = self.hdulist[0].header["BITPIX"]

            try:
                self.bzero = self.hdulist[0].header["BZERO"]
                self.bscale = self.hdulist[0].header["BSCALE"]
            except KeyError:
                self.bzero = 0
                self.bscale = 0

            self.data_type = self.hdulist[0].data.dtype
            self.bitpix2 = self.hdulist[0].header["BITPIX"]
            self.save_data_format = self.bitpix

        # get object name
        try:
            self.title = hdr["OBJECT"]
        except KeyError:
            self.title = ""

        # check AzCam header
        try:
            azcam_head = hdr["AZCAM-HEAD"]
            if azcam_head == "OK":
                self.azcam_header = 1
            else:
                self.azcam_header = 0
        except KeyError:
            self.azcam_header = 0

        # set Array type - output data type
        self.array_type = self.data_types[self.bitpix2]

        # check if first item is col bin or row bin
        try:
            self.focalplane.col_bin = hdr["CCDBIN1"]
            self.focalplane.row_bin = hdr["CCDBIN2"]
        except KeyError:
            self.focalplane.col_bin = 1
            self.focalplane.row_bin = 1

        self.focalplane.refpix1 = 0.0
        self.focalplane.refpix2 = 0.0

        if self.azcam_header == 1:
            # create empty arrays for focal plane values
            self.focalplane.amp_cfg = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.det_number = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.ext_number = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.jpg_ext = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.detpos_x = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.detpos_y = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.amppos_x = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.amppos_y = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.amppix1 = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.amppix2 = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.gapx = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.gapy = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.extpos_x = numpy.ndarray(shape=(cntExt), dtype="<u2")
            self.focalplane.extpos_y = numpy.ndarray(shape=(cntExt), dtype="<u2")

            self.focalplane.ext_name = cntExt * [""]

            # prepare arrays for image transformations
            self.focalplane.wcs.atm_1_1 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.atm_2_2 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.atv1 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.atv2 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.ltm_1_1 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.ltm_2_2 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.ltv_1 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.ltv_2 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.dtm_1_1 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.dtm_2_2 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.dtv_1 = numpy.ndarray(shape=(cntExt), dtype="<i2")
            self.focalplane.wcs.dtv_2 = numpy.ndarray(shape=(cntExt), dtype="<i2")

            # prepare arrays for WCS
            self.focalplane.wcs.rot_deg = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.scale1 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.scale2 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.cd_1_1 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.cd_1_2 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.cd_2_1 = numpy.ndarray(shape=(cntExt), dtype="float32")
            self.focalplane.wcs.cd_2_2 = numpy.ndarray(shape=(cntExt), dtype="float32")

            # read focal plane keywords from the main header
            try:
                self.focalplane.numdet_x = hdr["NUM-DETX"]
                self.focalplane.numdet_y = hdr["NUM-DETY"]
                self.focalplane.numamps_x = hdr["NUM-AMPX"]
                self.focalplane.numamps_y = hdr["NUM-AMPY"]
                self.focalplane.refpix1 = hdr["REF-PIX1"]
                self.focalplane.refpix2 = hdr["REF-PIX2"]

                # update num_ser_amps_det and num_par_amps_det
                self.focalplane.num_ser_amps_det = self.focalplane.numamps_x
                self.focalplane.num_par_amps_det = self.focalplane.numamps_y
            except KeyError:
                pass

            if NumExt == 0:
                try:
                    self.focalplane.amp_cfg[0] = hdr["AMP-CFG"]
                    self.focalplane.det_number[0] = hdr["DET-NUM"]
                    self.focalplane.ext_number[0] = hdr["EXT-NUM"]
                    self.focalplane.jpg_ext[0] = hdr["JPG-EXT"]
                    self.focalplane.detpos_x[0] = hdr["DET-POSX"]
                    self.focalplane.detpos_y[0] = hdr["DET-POSY"]
                    self.focalplane.amppos_x[0] = hdr["AMP-POSX"]
                    self.focalplane.amppos_y[0] = hdr["AMP-POSY"]
                    self.focalplane.amppix1[0] = hdr["AMP-PIX1"]
                    self.focalplane.amppix2[0] = hdr["AMP-PIX2"]
                except KeyError:
                    pass
        else:
            pass

        if NumExt == 0:
            # single extension file
            numcols = hdr["NAXIS1"]  # includes overscan
            numrows = hdr["NAXIS2"]
            numcols = int(numcols)
            numrows = int(numrows)

            # get image size
            self.size_x = numcols
            self.size_y = numrows

            self.focalplane.numcols_amp = numcols
            self.focalplane.numrows_amp = numrows
            self.focalplane.numpix_amp = numrows * numcols
            self.focalplane.numcols_image = numcols
            self.focalplane.numrows_image = numrows
            self.focalplane.numpixels = numrows * numcols

            # read overscan and prescan values
            try:
                self.focalplane.numcols_overscan = hdr["OVRSCAN1"]
                self.focalplane.numrows_overscan = hdr["OVRSCAN2"]
                self.focalplane.numcols_underscan = hdr["PRESCAN1"]
                self.focalplane.numrows_underscan = hdr["PRESCAN1"]
            except KeyError:
                pass

            # create offsets and scales arrays with default values
            self.offsets = numpy.empty(shape=[1], dtype="float32")
            self.scales = numpy.empty(shape=[1], dtype="float32")
            self.offsets[0] = 0.0
            self.scales[0] = 1.0

        else:
            # multiple extension file
            try:
                hdr = pyfits.getheader(filename, 1)
                section = hdr[
                    "DATASEC"
                ]  # includes overscan, total binned pixels per amp
                section = section.lstrip("[")
                section = section.split(":")
                fc = int(section[0])
                section1 = section[1].split(",")
                lc = int(section1[0])
                fr = int(section1[1])
                lr = int(section[2].rstrip("]"))
                numrows = lr - fr + 1
                numcols = lc - fc + 1

                self.focalplane.numviscols_amp = numcols
                self.focalplane.numvisrows_amp = numrows

                numcols = int(hdr["NAXIS1"])
                numrows = int(hdr["NAXIS2"])

                hdr = pyfits.getheader(filename, 0)
                self.focalplane.numamps_image = int(hdr["NAMPS"])
            except KeyError:
                pass

            # overscan and underscan taken from the first extension
            try:
                self.focalplane.numcols_overscan = self.hdulist[1].header["OVRSCAN1"]
                self.focalplane.numcols_underscan = self.hdulist[1].header["PRESCAN1"]
                self.focalplane.numrows_overscan = self.hdulist[1].header["OVRSCAN2"]
                self.focalplane.numrows_underscan = self.hdulist[1].header["PRESCAN2"]
            except KeyError:
                self.focalplane.numcols_overscan = 0
                self.focalplane.numcols_underscan = 0
                self.focalplane.numrows_overscan = 0
                self.focalplane.numrows_underscan = 0

            self.focalplane.num_amps_det = NumExt
            self.focalplane.numpix_amp = numrows * numcols
            self.focalplane.numcols_amp = numcols
            self.focalplane.numrows_amp = numrows

            self.focalplane.numcols_image = numcols * self.focalplane.num_ser_amps_det
            self.focalplane.numpixels = (
                self.focalplane.numpix_amp * self.focalplane.numamps_image
            )
            self.focalplane.numrows_image = numrows * self.focalplane.num_par_amps_det

            self.size_x = numcols * self.focalplane.num_ser_amps_det
            self.size_y = numrows * self.focalplane.num_par_amps_det

            # create offsets and scales arrays with default values
            self.offsets = numpy.empty(shape=[last_ext - 1], dtype="float32")
            self.scales = numpy.empty(shape=[last_ext - 1], dtype="float32")

            for indx in range(0, NumExt):
                self.offsets[indx] = 0.0
                self.scales[indx] = 1.0

            if self.azcam_header == 1:
                for indx in range(1, NumExt + 1):
                    try:
                        self.focalplane.amp_cfg[indx - 1] = self.hdulist[indx].header[
                            "AMP-CFG"
                        ]
                        self.focalplane.det_number[indx - 1] = self.hdulist[
                            indx
                        ].header["DET-NUM"]
                        self.focalplane.ext_number[indx - 1] = self.hdulist[
                            indx
                        ].header["EXT-NUM"]
                        self.focalplane.jpg_ext[indx - 1] = self.hdulist[indx].header[
                            "JPG-EXT"
                        ]
                        self.focalplane.detpos_x[indx - 1] = self.hdulist[indx].header[
                            "DET-POSX"
                        ]
                        self.focalplane.detpos_y[indx - 1] = self.hdulist[indx].header[
                            "DET-POSY"
                        ]
                        self.focalplane.extpos_x[indx - 1] = self.hdulist[indx].header[
                            "EXT-POSX"
                        ]
                        self.focalplane.extpos_y[indx - 1] = self.hdulist[indx].header[
                            "EXT-POSY"
                        ]
                        self.focalplane.amppix1[indx - 1] = self.hdulist[indx].header[
                            "AMP-PIX1"
                        ]
                        self.focalplane.amppix2[indx - 1] = self.hdulist[indx].header[
                            "AMP-PIX2"
                        ]
                        # self.focalplane.refpix1[indx - 1] = self.hdulist[indx].header["CRPIX1"]
                        # self.focalplane.refpix2[indx - 1] = self.hdulist[indx].header["CRPIX2"]

                        self.focalplane.ext_name[indx - 1] = f"IM{indx}"  # new

                        DetSec = self.hdulist[indx].header["DETSEC"]
                        DetSec = (DetSec.lstrip("[").rstrip("]")).split(",")

                        self.focalplane.gapx[indx - 1] = float(
                            self.focalplane.amppix1[indx - 1]
                        ) - float(DetSec[0].split(":")[0])
                        self.focalplane.gapy[indx - 1] = float(
                            self.focalplane.amppix2[indx - 1]
                        ) - float(DetSec[1].split(":")[0])

                        self.focalplane.amppos_x[indx - 1] = self.hdulist[indx].header[
                            "AMP-POSX"
                        ]
                        self.focalplane.amppos_y[indx - 1] = self.hdulist[indx].header[
                            "AMP-POSY"
                        ]

                    except KeyError:
                        pass

                    # read the WCS keywords from main header
                    try:
                        # image transformation keywords
                        self.focalplane.wcs.atm_1_1[indx - 1] = self.hdulist[
                            indx
                        ].header["ATM1_1"]
                        self.focalplane.wcs.atm_2_2[indx - 1] = self.hdulist[
                            indx
                        ].header["ATM2_2"]
                        self.focalplane.wcs.atv1[indx - 1] = self.hdulist[indx].header[
                            "ATV1"
                        ]
                        self.focalplane.wcs.ltv_2[indx - 1] = self.hdulist[indx].header[
                            "ATV2"
                        ]
                        self.focalplane.wcs.ltm_1_1[indx - 1] = self.hdulist[
                            indx
                        ].header["LTM1_1"]
                        self.focalplane.wcs.ltm_2_2[indx - 1] = self.hdulist[
                            indx
                        ].header["LTM2_2"]
                        self.focalplane.wcs.ltv_1[indx - 1] = self.hdulist[indx].header[
                            "LTV1"
                        ]
                        self.focalplane.wcs.ltv_2[indx - 1] = self.hdulist[indx].header[
                            "LTV2"
                        ]
                        self.focalplane.wcs.dtm_1_1[indx - 1] = self.hdulist[
                            indx
                        ].header["DTM1_1"]
                        self.focalplane.wcs.dtm_2_2[indx - 1] = self.hdulist[
                            indx
                        ].header["DTM2_2"]
                        self.focalplane.wcs.dtv_1[indx - 1] = self.hdulist[indx].header[
                            "DTV1"
                        ]
                        self.focalplane.wcs.dtv_2[indx - 1] = self.hdulist[indx].header[
                            "DTV2"
                        ]

                        # WCS keywords
                        self.focalplane.wcs.rot_deg[indx - 1] = self.hdulist[
                            indx
                        ].header["ROT-DEG"]
                        self.focalplane.wcs.scale1[indx - 1] = self.hdulist[
                            indx
                        ].header["SCALE1"]
                        self.focalplane.wcs.scale2[indx - 1] = self.hdulist[
                            indx
                        ].header["SCALE2"]
                        self.focalplane.wcs.cd_1_1[indx - 1] = self.hdulist[
                            indx
                        ].header["CD1_1"]
                        self.focalplane.wcs.cd_1_2[indx - 1] = self.hdulist[
                            indx
                        ].header["CD1_2"]
                        self.focalplane.wcs.cd_2_1[indx - 1] = self.hdulist[
                            indx
                        ].header["CD2_1"]
                        self.focalplane.wcs.cd_2_2[indx - 1] = self.hdulist[
                            indx
                        ].header["CD2_2"]

                    except KeyError:
                        pass

        # ---------------------------- data -------------------------------------------------------

        # create .data numpy array and scale data,
        #    .hdulist[0].data is [nrows][ncols] -> .data[0] is the first row
        if NumExt == 0:
            self.data = numpy.ndarray(
                shape=(1, self.focalplane.numpix_amp),
                buffer=self.hdulist[0].data,
                dtype=self.data_type,
            ).copy()
        else:
            self.data = numpy.empty(
                shape=[self.focalplane.numamps_image, self.focalplane.numpix_amp],
                dtype=self.data_type,
            ).copy()

            for chan in range(first_ext, last_ext):
                self.data[chan - 1, :] = numpy.ndarray(
                    shape=(self.focalplane.numpix_amp),
                    buffer=self.hdulist[chan].data,
                    dtype=self.data_type,
                ).copy()

        # take care of the big-endian/little-endian format
        if self.array_type == "float64":
            self.data = self.data.astype("float64")
        else:
            self.data = self.data.astype("float32")

        self.hdulist.close()

        self.is_valid = 1

        if self.array_type == "float64":
            self.buffer = numpy.empty(shape=[self.size_y, self.size_x], dtype="float64")

            self.in_buffer = numpy.empty(
                shape=[self.size_y, self.size_x], dtype="float64"
            )
            self.out_buffer = numpy.empty(
                shape=[self.size_y, self.size_x], dtype="float64"
            )
            self.in_buffer = self.data.astype("float64")
        else:
            self.buffer = numpy.empty(shape=[self.size_y, self.size_x], dtype="float32")

            self.in_buffer = numpy.empty(
                shape=[self.size_y, self.size_x], dtype="float32"
            )
            self.out_buffer = numpy.empty(
                shape=[self.size_y, self.size_x], dtype="float32"
            )
            self.in_buffer = self.data.astype("float32")

        # set flags
        self.from_file = 1
        self.is_written = 1

        return

    def _write_fits_file(self, filename, filetype=0):
        """
        Write the FITS or MEF image to disk
        filetype is 0, 1, or 6 for FITS, MEF, or assembled.
        """

        Overwrite = self.overwrite or self.test_image

        fldr = os.path.dirname(filename)
        fldr = os.path.normpath(fldr)
        if fldr != "" and not os.path.exists(fldr):
            s = f"folder {fldr} does not exist"
            raise FileNotFoundError(s)

        # ERROR if file exists and overwrite flag not set
        if os.path.exists(filename):
            if Overwrite:
                loop = 0
                while loop < 10:
                    try:
                        os.remove(filename)
                        break
                    except Exception as details:
                        s = "ERROR deleting previous image file: %s" % repr(details)
                        loop += 1
                        time.sleep(0.5)
                if loop > 10:
                    return ["ERROR", s]
            else:
                s = "ERROR " + filename + " exists but Overwrite flag is not set"
                return ["ERROR", s]

        # assemble image as needed
        if self.assemble_image:
            self.assemble()

        if filetype == self.filetypes["FITS"]:
            self._write_standardfits_file(filename)
        elif filetype == self.filetypes["MEF"]:
            self._write_mef_file(filename)
        elif filetype == self.filetypes["ASM"]:
            self._write_asm_fits_file(filename)

        return

    def _write_standardfits_file(self, filename):
        """
        Write a standard (non-MEF) FITS file.
        """

        # allow case sensitive ext_name
        pyfits.EXTENSION_NAME_CASE_SENSITIVE = True

        # make PHU with data
        if self.focalplane.numamps_image == 1:
            if self.transposed_image:
                nx = self.focalplane.numcols_image
                ny = self.focalplane.numrows_image
            else:
                nx = self.focalplane.numrows_image
                ny = self.focalplane.numcols_image
            data = numpy.ndarray(
                shape=(nx, ny),
                dtype=self.data_types[self.save_data_format],
                buffer=self.data[0],
            )
            if self.transposed_image:
                data = data.transpose()
            if self.flip_image:
                data = numpy.flipud(data)
            hdu = pyfits.PrimaryHDU(data=data)
        else:
            if not self.assembled:
                self.assemble(1)
            hdu = pyfits.PrimaryHDU(data=self.buffer)

        # add header cards to PHU
        hdu.header.set("NAXIS", 2, "number of data axes")
        self._write_PHU(hdu)

        # update focal plane keywords
        self.focalplane.update_header_keywords()
        self.focalplane.update_ext_keywords()

        # add coord header cards to PHU
        self._write_extension_header(1, hdu)

        # add WCS header cards
        self._write_wcs_keywords(1, hdu)

        # create a hdu list
        self.hdulist = pyfits.HDUList([hdu])

        # write it all to a disk file (mod 11jul13)
        try:
            self.hdulist.writeto(filename)
        except Exception:
            time.sleep(0.2)
            self.hdulist.writeto(filename)

        self.hdulist.close()

        return

    def _write_mef_file(self, filename):
        """
        Write an MEF image file.
        """

        # allow case sensitive ext_name
        pyfits.EXTENSION_NAME_CASE_SENSITIVE = True

        # make PHU (no data)
        phdu = pyfits.PrimaryHDU()

        numHDUs = self.focalplane.numamps_image

        # add header cards to PHU
        self._write_PHU(phdu)

        # create a list of hdu's for MEF file
        self.hdulist = pyfits.HDUList([phdu])

        # loop through HDU's, creating extensions and writing data
        for ext_number in range(1, numHDUs + 1):  # first HDU is 1 not 0
            # create the extension name
            ext_name = self.focalplane.ext_name[ext_number - 1]

            # get the image data for this extension
            numrows_amp = self.focalplane.numrows_amp
            numcols_amp = self.focalplane.numcols_amp
            data = numpy.ndarray(
                shape=(numrows_amp, numcols_amp),
                dtype=self.data_types[self.save_data_format],
                buffer=self.data[ext_number - 1],
            )

            hdu = pyfits.ImageHDU(data=data, name=str(ext_name))
            hdu.header.set("NAXIS", 2, "number of data axes")
            hdu.header.set("INHERIT", True, "extension inherits PHDU keyword/values?")
            hdu.header.set("BUNIT", "ADU", "Physical unit of array values")

            # add coord header cards to this extension
            self._write_extension_header(ext_number, hdu)

            # add WCS header cards to this extension
            self._write_wcs_keywords(ext_number, hdu)

            # add Focal plane header cards to this extension
            self._write_focalplane_keywords(ext_number, hdu)

            # hdu is read-only, so make a copy in order to scale(I have no idea why)
            hdu1 = hdu.copy()
            hdu1.scale("int16", "", bzero=32768, bscale=1)

            # keywords may be removed so make sure and replace, ???
            try:
                del hdu1.header["BZERO"]
                del hdu1.header["BSCALE"]
            except KeyError:
                pass
            hdu1.header.set("BZERO", 32768.0, after=7)
            hdu1.header.set("BSCALE", 1.0, after=8)

            # append to hdulist
            self.hdulist.append(hdu1)

        # now write it all to a disk file
        self.hdulist.writeto(filename)

        self.hdulist.close()

        return

    def _write_asm_fits_file(self, filename):
        """
        Write an assembled single extension FITS image file.
        """

        # allow case sensitive ext_name
        pyfits.EXTENSION_NAME_CASE_SENSITIVE = True

        # make PHU with data
        if not self.assembled:
            self.assemble(1)

        data = numpy.ndarray(
            shape=(self.asmsize[1], self.asmsize[0]),
            dtype=self.data_types[self.save_data_format],
            buffer=self.buffer.astype(self.data_types[self.save_data_format]),
        )
        hdu = pyfits.PrimaryHDU(data=data, header=self.hdulist[0].header)

        # add header cards to PHU
        hdu.header.set("NAXIS", 1, "number of data axes")
        self._write_asm_fits_header(hdu)

        # update focal plane keywords
        self.focalplane.update_header_keywords()

        # update focal plane extension keywords
        self.focalplane.update_ext_keywords()

        # create a new hdu list - do not overwrite the original header
        self.asm_header.hdulist = pyfits.HDUList([hdu])

        # write it all to a disk file
        if filename.startswith("!"):
            filename = filename[1:]
            overwrite = True
        else:
            overwrite = False
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.asm_header.hdulist.writeto(filename, overwrite=overwrite)

        self.asm_header.hdulist.close()

        return

    def _write_bin_file(self, filename):
        """
        Write buffer as binary image to disk.
        """

        self.size_x = self.focalplane.numcols_image
        self.size_y = self.focalplane.numrows_image

        if not self.assembled:
            self.assemble(1)

        with open(filename, "wb") as fd:
            fd.write(self.buffer.astype("uint16").squeeze())

        return



import socket
import time

import numpy

import azcam
import azcam.exceptions


class ReceiveData(object):
    """
    Exposure subclass to receive image data.
    """

    def __init__(self, exposure):

        self.exposure = exposure  # upper level exposure object

        self.DeinterlaceParams = ""
        self.port = 0
        self.is_valid = 1

        self.PixelsReadout = 0
        self.pixels_remaining = 0
        self.camserver = 0
        self.socket = 0

        # Deinterlace mode
        self.deinterlace_mode = 1
        # Number of pixels per amplifier
        self.numpix_amp = 0
        # Number of amplifiers
        self.numamps_image = 0

        # using this helps writing efficiency, bytes
        self.RecBufferSize = 5 * 1024 * 1024

    def receive_image_data(self, data_size):
        """
        Receive binary image data from controller server.
        data_size is bytes.
        """

        if azcam.db.tools["controller"].camserver.demo_mode:
            self.mock_data()
            return

        # create a new socket for binary data and connect to the controller server
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.socket.connect(
            (
                azcam.db.tools["controller"].camserver.host,
                azcam.db.tools["controller"].camserver.port,
            )
        )

        azcam.log(f"Receiving image data: {data_size} bytes", level=3)

        # Init Deinterlace
        self.numamps_image = self.exposure.image.focalplane.numamps_image
        self.numpix_amp = self.exposure.image.focalplane.numpix_amp

        reqCnt = min(
            data_size - 17, self.RecBufferSize - 17
        )  # 17 bytes for the data frame size (%16d + space)
        dataCnt = 0  # receved data counter
        repCnt = 0  # repeate data request counter: 100
        getData = b""
        totalpixels = int(data_size / 2)
        self.PixelsReadout = 0
        self.pixels_remaining = totalpixels

        # create temporary image buffer
        BufferTemp = numpy.empty(shape=(self.exposure.image.data.size), dtype="<u2")

        # set image data pointer
        ptrData = 0

        # loop over data just read, long repeat as images could be slow to start
        while (dataCnt < data_size) and (repCnt < 50):

            # check if aborted by user (from abort() - controller.abort()
            if (
                azcam.db.tools["exposure"].exposure_flag
                == azcam.db.tools["exposure"].exposureflags["ABORT"]
            ):

                # if in a sequence then let this readout finish
                if self.exposure.is_exposure_sequence:
                    pass  # return will not be an error
                else:
                    # break out of read loop
                    azcam.db.tools[
                        "controller"
                    ].readout_abort()  # stop ControllerServer
                    break

            getData = self.request_data(
                reqCnt + 17
            )  # request data + 17 bytes for data length
            len1 = len(getData)
            azcam.log(f"Readout: {self.pixels_remaining:10d} pixels remaining", level=3)

            if len1 != 0:
                dataCnt += len1
                repCnt = 0

                # store data
                pixelsreadout = int(
                    len1 / 2
                )  # number pixels in this read now available

                # convert received data to unsigned shorts
                ImageBufferTemp = numpy.ndarray(
                    shape=(1, pixelsreadout), dtype="<u2", buffer=getData
                )

                # copy the data into TempBuffer
                BufferTemp[ptrData : ptrData + pixelsreadout] = ImageBufferTemp[
                    0:pixelsreadout
                ]
                ptrData = ptrData + pixelsreadout

                reqCnt = min(data_size - dataCnt - 17, self.RecBufferSize - 17)
                self.PixelsReadout = self.PixelsReadout + pixelsreadout
                self.pixels_remaining = self.pixels_remaining - pixelsreadout
                # time.sleep(0.2)
            else:
                time.sleep(0.2)
                repCnt = repCnt + 1

        # check if all data has been received
        if dataCnt == data_size:
            self.is_valid = 1
            self.pixels_remaining = 0
            azcam.log("Image data received")
        else:
            if (
                not azcam.db.tools["exposure"].exposure_flag
                == azcam.db.tools["exposure"].exposureflags["ABORT"]
            ):
                s = "ERROR in ReceiveImageData: Received %d of %d bytes" % (
                    dataCnt,
                    data_size,
                )
                self.socket.close()
                raise azcam.exceptions.AzcamError(s)
            else:
                self.socket.close()
                raise azcam.exceptions.AzcamError(
                    "Aborted in receive_image_data", error_code=3
                )
        self.socket.close()

        # deinterlace into exposure.image.data
        BufferTemp = BufferTemp.reshape(self.numpix_amp, self.numamps_image)

        if len(self.exposure.data_order) == 0:
            for AmpPos in range(0, self.numamps_image):
                self.exposure.image.data[AmpPos, 0 : self.numpix_amp] = BufferTemp[
                    0 : self.numpix_amp, AmpPos
                ]
        else:
            indx = 0
            for item in self.exposure.data_order:
                self.exposure.image.data[indx, 0 : self.numpix_amp] = BufferTemp[
                    0 : self.numpix_amp, item
                ]
                indx += 1

        return

    def request_data(self, datacnt):

        request = "GetImageData " + str(datacnt) + "\n"
        self.socket.send(str.encode(request))

        loop = 1
        rptCnt = 10
        data = b""
        gotCnt = 0
        dataFrame = b""
        oneFrame = b""
        cntFrame = 0

        datacnt += 17
        while (loop == 1) and (rptCnt > 0):
            oneFrame = self.socket.recv(datacnt)
            if len(oneFrame) > 0:
                dataFrame += oneFrame
                cntFrame += 1

                if gotCnt == 0:
                    if len(dataFrame) >= 17:
                        dataCnt = int(dataFrame[0:16])
                        if dataCnt > 0:
                            data += dataFrame[17:]
                            if dataCnt == len(dataFrame[17:]):
                                gotCnt = 0
                                loop = 0
                            else:
                                dataFrame = dataFrame[17:]
                                datacnt = datacnt - 17 - len(dataFrame[17:])
                                gotCnt = 1
                        else:
                            loop = 0
                            data = ""
                    else:
                        datacnt = datacnt - len(dataFrame)
                        rptCnt -= 1
                else:
                    if dataCnt == len(dataFrame):
                        data = dataFrame
                        loop = 0

            else:  # time out: received no data
                rptCnt -= 1

        if rptCnt == 0:
            data = ""

        return data

    def mock_data(self):
        """
        Generate mock data for demo mode.
        """

        ix = self.exposure.image.focalplane.numcols_image
        iy = self.exposure.image.focalplane.numrows_image

        for AmpPos in range(azcam.db.tools["exposure"].image.focalplane.numamps_image):
            self.exposure.image.data[AmpPos] = numpy.linspace(
                0, 65355, int(iy * ix / self.image.focalplane.numamps_image)
            )

        return



"""
API for azcam
"""

import typing

import azcam


class API(object):
    """
    AzCam API class.
    """

    def __init__(self):

        self.mock = 0
        """True for mock mode which uses no hardware"""

    def initialize(self, mock: int = 0):
        """
        Initialize the API.

        Args:
            mock: flag set true for development with no hardware

        """

        self.mock = mock
        self.parameters = azcam.db.parameters

        try:
            self.exposure = azcam.db.tools["exposure"]
        except Exception:
            pass
        try:
            self.tempcon = azcam.db.tools["tempcon"]
        except Exception:
            pass
        try:
            self.instrument = azcam.db.tools["instrument"]
        except Exception:
            pass
        try:
            self.telescope = azcam.db.tools["telescope"]
        except Exception:
            pass
        try:
            self.controller = azcam.db.tools["controller"]
        except Exception:
            pass
        try:
            self.focus = azcam.db.tools["focus"]
        except Exception:
            pass

        self.is_initialized = 1

        return

    # *************************************************************************
    #   parameters
    # *************************************************************************

    def get_par(self, parameter: str, subdict: str | None = None) -> typing.Any:
        """
        Return the current attribute value of a parameter in the parameters dictionary.
        If subdict is not specified then the default sub-dictionary is used.

        Args:
            parameter: name of the parameter
            subdict: name of the sub-dictionary containing the parameter

        Returns:
            value: value of the parameter
        """

        return self.parameters.get_par(parameter, subdict)

    def set_par(
        self, parameter: str, value: typing.Any = "None", subdict: str | None = None
    ) -> None:
        """
        Set the value of a parameter in a parameters dictionary.
        Also attempts to set actual attribute value.

        Args:
            parameter: name of the parameter
            value: value of the parameter. Defaults to None.
            subdict: name of sub-dictionary in which to set paramater
        """

        return self.parameters.set_par(parameter, value, subdict)

    def save_pars(self) -> None:
        """
        Writes the par_dict to the par_file using current values.
        """

        return self.parameters.save_pars()

    # *************************************************************************
    #   exposures
    # *************************************************************************

    def expose(self, exposure_time: float, imagetype: str = "", title: str = ""):
        """
        Make a complete exposure.

        Args:
            exposure_time: exposure time in seconds
            imagetype: type of exposure ('zero', 'object', 'flat', ...)
            title: image title.
        """

        return self.exposure.expose(exposure_time, imagetype, title)

    def expose1(
        self, exposure_time: float, image_type: str = "", image_title: str = ""
    ):
        """
        Make a complete exposure with immediate return to caller.

        Args:
            exposure_time: exposure time in seconds
            image_type: type of exposure ('zero', 'object', 'flat', ...)
            image_title: image title, usually surrounded by double quotes
        """

        return self.exposure.expose1(exposure_time, image_type, image_title)

    def sequence(
        self, number_exposures: int = 1, flush_array_flag: int = -1, delay: float = -1
    ):
        """
        Take an exposure sequence.
        Uses pre-set exposure time, image type and image title.

        Args:
            number_exposures: number of exposures to make.
            flush_array_flag: defines detector flushing:
                -1 => current value defined by exposure.exposure_sequence_flush [default]
                0 => flush for each exposure
                1 => flush after first exposure only
                2 => no flush
            delay: delay between exposures in seconds
                -1 => no change
        """

        return self.exposure.sequence(number_exposures, flush_array_flag, delay)

    def sequence1(
        self, number_exposures: int = 1, flush_array_flag: int = -1, delay: float = -1
    ):
        """
        Take an exposure sequence.
        Uses pre-set exposure time, image type and image title.

        Args:
            number_exposures: number of exposures to make.
            flush_array_flag: defines detector flushing:
                -1 => current value defined by exposure.exposure_sequence_flush [default]
                0 => flush for each exposure
                1 => flush after first exposure only
                2 => no flush
            delay: delay between exposures in seconds
                -1 => no change
        """

        return self.exposure.sequence1(number_exposures, flush_array_flag, delay)

    def test(self, exposure_time: float = 0.0, shutter: int = 0):
        """
        Make a test exposure.

        Args:
            exposure_time: exposure time in seconds
            shutter: 0 for closed and 1 for open
        """

        return self.exposure.test(exposure_time, shutter)

    def flush(self, cycles: int = 1):
        """
        Flush/clear sensor.

        Args:
            cycles: number of times to flush the sensor
        """

        return self.exposure.flush(cycles)

    def set_filename(self, filename: str):
        """
        Set the filename components based on a simple filename.

        Args:
            filename: complete filename to be set
        """

        return self.exposure.set_filename(filename)

    def initialize_exposure(self):
        """
        Initialize exposure.
        """

        return self.exposure.initialize()

    def reset_exposure(self):
        """
        Reset exposure.
        """

        return self.exposure.reset()

    def get_roi(self, roi_num: int = 0) -> list:
        """
        Returns a list of the ROI parameters for the roi_num specified.
        Currently only one ROI (0) is supported.

        Args:
            roi_num: ROI number to return

        Returns: list format is (first_col,last_col,first_row,last_row,col_bin,row_bin).
        """

        return self.exposure.get_roi(roi_num)

    def set_roi(
        self,
        first_col: int = -1,
        last_col: int = -1,
        first_row: int = -1,
        last_row: int = -1,
        col_bin: int = -1,
        row_bin: int = -1,
        roi_num: int = 0,
    ):
        """
        Sets the ROI values for subsequent exposures.
        Currently only one ROI [0] is supported.
        These values are for the entire focal plane, not just one detector.
        All values are in unbinned coordinates.

        Args:
            first_col: first column
            last_col: last  column
            first_row: first row
            last_row: last row
            col_bin: : column binning
            row_bin: row binning
            roi_num: ROI number [0]
        """

        return self.exposure.set_roi(
            first_col, last_col, first_row, last_row, col_bin, row_bin, roi_num
        )

    def roi_reset(self):
        """
        Resets detector ROI values to full frame, current binning.
        """

        return self.exposure.roi_reset()

    def get_format(self) -> list:
        """
        Return the current detector format parameters.
        """

        return self.exposure.get_format()

    def set_format(
        self,
        ns_total: int,
        ns_predark: int,
        ns_underscan: int,
        ns_overscan: int,
        np_total: int,
        np_predark: int,
        np_underscan: int,
        np_overscan: int,
        np_frametransfer: int,
    ) -> None:
        """
        Set the sensor format for subsequent exposures.
        Must call set_roi() after using this command and before starting exposure.

        Args:
            ns_total: number of visible columns.
            ns_predark: number of physical dark underscan columns.
            ns_underscan: desired number of desired dark underscan columns.
            ns_overscan: number of dark overscan columns.
            np_total: number of visible rows.
            np_predark: number of physical dark underscan rows.
            np_underscan: number of desired dark underscan rows.
            np_overscan: number of desired dark overscan rows.
            np_frametransfer: number rows to frame transfer shift.
        """

        return self.exposure.set_format(
            ns_total,
            ns_predark,
            ns_underscan,
            ns_overscan,
            np_total,
            np_predark,
            np_underscan,
            np_overscan,
            np_frametransfer,
        )

    def abort(self):
        """
        Abort an operation in progress.
        """

        return self.exposure.abort()

    def pause(self):
        """
        Pause an exposure inegration (only) in progress.
        """

        return self.exposure.pause()

    def resume(self):
        """
        Resume a paused exposure.
        """

        return self.exposure.resume()

    def set_shutter(self, state: int = 0, shutter_id: int = 0):
        """
        Open or close a shutter.

        :param state:
        :param shutter_id: Shutter ID flag

          * 0 => controller default shutter.
          * 1 => instrument default shutter.
        """

        state = int(state)
        shutter_id = int(shutter_id)

        return self.exposure.set_shutter(state, shutter_id)

    def read_header_file(self, filename):
        """
        Read header file located on the server machine.
        """

        return self.exposure.read_header_file(filename)

    def get_exposuretime(self) -> float:
        """
        Return current exposure time in seconds.

        Returns:
            exposure_time: exposure time in seconds.
        """

        return self.exposure.get_exposuretime()

    def set_exposuretime(self, exposure_time: float):
        """
        Set current exposure time.

        Args:
            exposure_time: exposure time in seconds.
        """

        return self.exposure.set_exposuretime(exposure_time)

    def get_exposuretime_remaining(self) -> float:
        """
        Return remaining exposure time in seconds.

        Returns:
            exposure_time_remaining: exposure time remaining in seconds.
        """

        return self.exposure.get_exposuretime_remaining()

    def get_pixels_remaining(self) -> int:
        """
        Return number of remaining pixels to be read (counts down).

        Returns:
            pixels_remaining: number of pixels remaining to be readout.
        """

        return self.exposure.get_pixels_remaining()

    def get_status(self) -> dict:
        """
        Return a variety of system status data in one dictionary.

        Returns:
            data: dictionary of exposure related data
        """

        return azcam.db.tools["exposure"].get_status()

    def set_image_title(self, title: str):
        """
        Set the image title.

        Args:
            title: image title
        """

        return self.exposure.set_image_title(title)

    def get_image_title(self):
        """
        Return the image title.
        """

        return self.exposure.get_image_title()

    def set_image_type(self, imagetype: str = "zero"):
        """
        Set image type for an exposure.

        Args:
            imagetype: system defined, and typically includes: zero, object, dark, flat.
        """

        return self.exposure.set_image_type(imagetype)

    def get_image_type(self) -> str:
        """
        Get current image type for an exposure.
        imagetype is system defined, and typically includes:
        zero, object, dark, flat.

        Returns:
            image_type: image type string
        """

        return self.exposure.get_image_type()

    def get_image_types(self) -> list[str]:
        """
        Return a list of valid imagetypes.

        Returns:
            image_types: list of valid iamge types
        """

        return self.exposure.get_image_types()

    # *************************************************************************
    #   temperatures
    # *************************************************************************

    def get_temperatures(self) -> list[float]:
        """
        Return all system temperatures.

        Returns:
            temperatures: list of temperatures read
        """

        return self.tempcon.get_temperatures()

    def set_control_temperature(
        self, temperature: float | None = None, temperature_id: int = -1
    ) -> None:
        """
        Set the control temperature (set point).

        Args:
            temperature: control temperature in Celsius. If not specified, use saved value
            temperature_id: control temperature sensor number
        """

        return self.tempcon.set_control_temperature(temperature, temperature_id)

    def get_control_temperature(self, temperature_id: int = -1) -> float:
        """
        Get the control temperature (set point).

        Args:
            temperature_id: temperature sensor identifier

        Returns:
            control_temperature: control temperature
        """

        return self.tempcon.get_control_temperature(temperature_id)

    # *************************************************************************
    #   instruments
    # *************************************************************************

    def get_filters(self, filter_id=0):
        """
        Return a list of all available/loaded filters.
        Args:
            filter_id: filter mechanism ID
        """

        return self.instrument.get_filters(filter_id)

    def get_filter(self, filter_id=0):
        """
        Return the current/loaded filter, typically the filter in the beam.
        Args:
            filter_id: filter mechanism ID
        """

        return self.instrument.get_filter(filter_id)

    def set_filter(self, filter_name, filter_id=0):
        """
        Set the current/loaded filter, typically the filter in the beam.
        Args:
            filter_name: filter name to set. Could be a number or filter name.
            filter_id: filter mechanism ID
        """

        return self.instrument.set_filter(filter_name, filter_id)

    def get_wavelengths(self, wavelength_id: int = 0):
        """
        Returns a list of valid wavelengths.
        Used for filter and LED based systems.
        Args:
            wavelength_id: wavelength mechanism ID
        """

        return self.instrument.get_wavelengths(wavelength_id)

    def get_wavelength(self, wavelength_id: int = 0):
        """
        Returns the current wavelength.
        Args:
            wavelength_id: wavelength mechanism ID
        """

        return self.instrument.get_wavelength(wavelength_id)

    def set_wavelength(self, wavelength: typing.Any, wavelength_id: int = 0):
        """
        Set the current wavelength, typically for a filter or grating.
        Args:
            wavelength: wavelength value to set. Could be a number or filter name.
            wavelength_id: wavelength mechanism ID
        """

        return self.instrument.set_wavelength(wavelength, wavelength_id)

    # *************************************************************************
    #   focus
    # *************************************************************************
    def focus_initalize(self):
        """
        Initialize focus routine.
        """

        return self.focus.initialize()

    def focus_abort(self):
        """
        Abort focus exposure.
        """

        return self.focus.abort()

    def focus_set_pars(
        self,
        exposure_time: float,
        number_exposures: int = 7,
        focus_step: float = 30,
        detector_shift: int = 10,
    ):
        """
        Set focus related parameters.
        Args:
            number_exposures: Number of exposures in focus sequence.
            focus_step: Number of focus steps between each exposure in a frame.
            detector_shift: Number of rows to shift detector for each focus step.
            exposuretime: Exposure time i seconds.

        """

        return self.focus.set_pars(
            exposure_time,
            number_exposures,
            focus_step,
            detector_shift,
        )

    def focus_run(
        self,
        exposure_time: float,
        number_exposures: int,
        focus_step: float,
        detector_shift: int,
    ):
        """
        Execute the focus sequence.
        If focus.set_pars() was previously called then those values are used and input here is ignored.
        Args:
            exposure_time: Exposure time in seconds.
            number_exposures: Number of exposures in focus sequence.
            focus_step: Number of focus steps between each exposure in a frame.
            detector_shift: Number of rows to shift detector for each focus step.
        """

        return self.focus.run(
            exposure_time, number_exposures, focus_step, detector_shift
        )



"""
Contains the Image class.
"""

import os

import azcam
import azcam.utils
import azcam.exceptions
import numpy
from azcam.header import Header
from azcam.image_focalplane import FocalPlane
from azcam.image_headers import ImageHeaders
from azcam.image_io import ImageIO


class Image(ImageHeaders, ImageIO):
    """
    Class to create and manipulate the standard azcam image object.
    """

    def __init__(self, filename=""):
        super().__init__()

        self.is_valid = 0
        """True when image is valid"""
        self.is_written = 0
        """True when image has been written to disk"""
        self.toggle = 0
        """True when image is first ready"""
        self.filename = ""
        """image filename"""
        self.overwrite = 0
        """True to allow overwritting image file"""
        self.test_image = 0
        """True when the image is a test image (overwrite is automatic)"""
        self.make_lockfile = 0
        """ True to make a lock file when image is written"""
        self.filetype = 0
        """image file type"""
        self.title = ""
        """title string"""

        # image size - columns
        self.size_x = 0
        # image szie - rows
        self.size_y = 0
        # numpy image buffer for assembled image [y,x]
        self.buffer = []
        self.in_buffer = []
        self.out_buffer = []
        # True if image was read from a file
        self.from_file = 0
        # True if image was read from a file and has AzCam header
        self.azcam_header = 0

        self.transposed_image = 0
        self.flip_image = 0

        # WCS
        self.write_wcs = 1
        # set default values for the scale and offset
        self.scales = []
        self.offsets = []
        # numpy image data buffer
        self.data = []
        # display image
        self.display_image = 0

        # assembly
        self.assemble_image = 0
        # flag to trim overscan, True means trim
        self.trim = 1
        # True when image Data has been assembled into Buffer
        self.assembled = 0
        # True if image is trimmed
        self.trimmed = 0
        # assembled image size (may be different due to trimming the prescan and overscan)
        self.asmsize = (0, 0)

        # Data type (numpy array data type after reading) - default 16-bit integer
        self.data_type = 16
        # BITPIX value - before accessing data buffer
        self.bitpix = 16
        # BITPIX value - after accessing data buffer
        self.bitpix2 = 0
        # BZERO value
        self.bzero = 0
        # BSCALE value
        self.bscale = 0

        # data types for fits images
        self.data_types = {
            8: "uint8",
            16: "uint16",
            32: "uint32",
            64: "uint64",
            -32: "float32",
            -64: "float64",
        }
        self.filetypes = {"FITS": 0, "MEF": 1, "BIN": 2, "ASM": 6}
        # final data array type
        self.array_type = 0
        # Allows saving data using other data format than BITPIX2
        self.save_data_format = 16

        # sub-tools
        self.header = Header()
        self.focalplane = FocalPlane()
        self.asm_header = Header()  # Header for assembled image

        # read a file if specified when instance created
        if filename != "":
            self.read_file(filename)

    def read_file(self, filename: str):
        """
        Read FITS image file (standard or MEF).
        """

        filename = azcam.utils.make_image_filename(filename)
        self.filename = filename

        self.filetype = self.filetypes["FITS"]
        self.assembled = 0
        self.is_valid = 0
        self.toggle = 0

        self.header = Header(self)  # new header

        # read file
        self._read_fits_file(self.filename)

        return

    def write_file(self, filename: str, filetype: int = -1):
        """
        Write image to disk file.
        filetype is 0 for FITS, 1 for MEF, 2 for BIN, 6 for assembled.
        """

        filename = azcam.utils.make_image_filename(filename)
        self.filename = filename

        # delete file if it exists
        if self.overwrite and os.path.exists(filename):
            os.remove(filename)

        if self.test_image and os.path.exists(filename):
            os.remove(filename)

        if filetype == -1:
            filetype = self.filetype

        if filetype == 0:
            self._write_fits_file(filename)
        elif filetype == 1:
            self._write_mef_file(filename)
        elif filetype == 2:
            self._write_bin_file(filename)
        elif filetype == 6:
            self._write_asm_fits_file(filename)
        else:
            raise azcam.exceptions.AzcamError("Invalid filetype for Image")

        # optionally make a lock file indicating the image file has been written
        if self.make_lockfile:
            lockfile = filename.replace(".bin", ".OK")
            with open(lockfile, "w"):
                pass

        return

    def assemble(self, trim: int = -1):
        """
        Assemble .data into .buffer.
        """

        if not self.is_valid:
            raise azcam.exceptions.AzcamError("image is not valid")

        if self.assembled:
            return

        # update self.asmsize
        self.asmsize = (self.focalplane.numcols_image, self.focalplane.numrows_image)

        # create self.buffer which is a single numpy image buffer for a fully assembled
        # image [row,cols] or [y,x]) - if image is not read from file
        if self.from_file != 1:
            if self.data.dtype == "float64":
                self.buffer = numpy.empty(
                    shape=[self.size_y, self.size_x], dtype="float64"
                )
            else:
                self.buffer = numpy.empty(
                    shape=[self.size_y, self.size_x], dtype="float32"
                )

        Offsets = self.offsets
        Scales = self.scales

        if trim == 1:
            prescan1 = self.focalplane.numcols_underscan
            overscan1 = self.focalplane.numcols_overscan
            prescan2 = self.focalplane.numrows_underscan
            overscan2 = self.focalplane.numrows_overscan

            # update the assembled image size

            # 12Sep13 Zareba
            num_under = self.focalplane.numamps_x * self.focalplane.numcols_overscan
            num_over = self.focalplane.numamps_x * self.focalplane.numcols_underscan
            size_x = self.size_x - num_under - num_over

            # 12Sep13 Zareba
            num_under = self.focalplane.numamps_y * self.focalplane.numrows_overscan
            num_over = self.focalplane.numamps_y * self.focalplane.numrows_underscan
            size_y = self.size_y - num_under - num_over

            self.asmsize = (size_x, size_y)
            self.asmsize = self.asmsize
            imgSize = size_x * size_y

            # reshape Buffer
            self.buffer = numpy.resize(self.buffer, imgSize)
            self.buffer = self.buffer.reshape((self.asmsize[1], self.asmsize[0]))

        else:
            prescan1 = 0
            overscan1 = 0
            prescan2 = 0
            overscan2 = 0

            self.asmsize = (self.size_x, self.size_y)
            self.asmsize = self.asmsize
            imgSize = self.size_x * self.size_y

            # reschape Buffer - if needed
            if self.data.size != self.buffer.size:
                self.buffer = numpy.resize(self.buffer, imgSize)
                self.buffer = self.buffer.reshape((self.asmsize[1], self.asmsize[0]))

        # destination AmpX and AmpY size corrected for prescan and overscan values
        dstAmpX = self.focalplane.numcols_amp - prescan1 - overscan1
        dstAmpY = self.focalplane.numrows_amp - prescan2 - overscan2

        # source AmpX and AmpY size including prescan and overscan
        srcAmpX = self.focalplane.numcols_amp
        srcAmpY = self.focalplane.numrows_amp

        ampX = self.focalplane.numcols_amp
        ampY = self.focalplane.numrows_amp

        self.lineLen = ampX - prescan1 - overscan1

        self.startLine = prescan2
        self.stopLine = ampY - prescan2 - overscan2

        Ext = self.focalplane.jpg_ext
        AmpFlip = self.focalplane.amp_cfg

        pixNum = 0

        for parAmps in range(0, self.focalplane.num_par_amps_det):
            # remove the prescan and overscane lines from the image
            extBase = parAmps * self.focalplane.num_ser_amps_det

            srcLine = prescan2
            for line in range(parAmps * dstAmpY, parAmps * dstAmpY + dstAmpY):
                lineStart = 0

                for currExt in range(
                    extBase, extBase + self.focalplane.num_ser_amps_det
                ):
                    # copy one line from the current extension

                    indx = Ext[currExt] - 1  # current amplifier
                    flip = int(
                        AmpFlip[indx]
                    )  # determine flip for the current extension

                    if flip == 0:  # no flip
                        posX = srcLine * srcAmpX + prescan1
                        self.buffer[line][lineStart : lineStart + dstAmpX] = (
                            self.data[indx][posX : posX + dstAmpX] - Offsets[indx]
                        ) * Scales[indx]

                        lineStart += self.lineLen
                        pixNum += self.lineLen

                    if flip == 1:  # x flip: reverse the readout sequence
                        posX = srcLine * srcAmpX + prescan1
                        self.buffer[line][lineStart : lineStart + dstAmpX] = (
                            self.data[indx][posX : posX + dstAmpX][::-1] - Offsets[indx]
                        ) * Scales[indx]

                        lineStart += self.lineLen
                        pixNum += self.lineLen

                    if flip == 2:  # y flip: get the flip line
                        posX = (srcAmpY - srcLine - overscan2 - 1) * srcAmpX + prescan1
                        self.buffer[line][lineStart : lineStart + dstAmpX] = (
                            self.data[indx][posX : posX + dstAmpX] - Offsets[indx]
                        ) * Scales[indx]

                        lineStart += self.lineLen
                        pixNum += self.lineLen

                    if (
                        flip == 3
                    ):  # xy flip; get the flip line and reverse the readout sequence
                        posX = (srcAmpY - srcLine - prescan2 - 1) * srcAmpX + prescan1
                        self.buffer[line][lineStart : lineStart + dstAmpX] = (
                            self.data[indx][posX : posX + dstAmpX][::-1] - Offsets[indx]
                        ) * Scales[indx]

                        lineStart += self.lineLen
                        pixNum += self.lineLen

                srcLine += 1

        # reshape the Buffer to 2D
        self.buffer = self.buffer.reshape((self.asmsize[1], self.asmsize[0]))

        # set isAssembled
        self.assembled = 1

        # set isTrimmed
        if trim == 1:
            self.trimmed = 1

        return

    def set_scaling(
        self, gains: list[float] | None = None, offsets: list[float] | None = None
    ):
        """
        Set gains and offsets for image assembly.

        Args:
            gains: gains values for each image section in [e/DN]
            offsets: offsets or bias values for each image section
        """

        self.num_extensions = self.focalplane.numamps_image

        # set default values for the scale and offset
        self.scales = numpy.empty(shape=[self.num_extensions], dtype="f")
        for ext in range(self.num_extensions):
            self.scales[ext] = 1.0
        self.offsets = numpy.empty(shape=[self.num_extensions], dtype="f")
        for ext in range(self.num_extensions):
            self.offsets[ext] = 0.0

        if gains is None:
            gains = len(self.data) * [1.0]

        if offsets is None:
            offsets = len(self.data) * [0.0]

        # Scales is gain (inverse electrical gain)
        for chan in range(len(self.data)):
            self.scales[chan] = gains[chan]
            self.offsets[chan] = offsets[chan]

        return
